import itertools
import logging
import typing
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pandas
import pendulum
import pyarrow

from tecton_core import feature_set_config
from tecton_core import specs
from tecton_core.arrow import arrow_to_pandas_dataframe
from tecton_core.data_types import BoolType
from tecton_core.data_types import Float32Type
from tecton_core.data_types import Float64Type
from tecton_core.data_types import Int32Type
from tecton_core.data_types import Int64Type
from tecton_core.data_types import StringType
from tecton_core.data_types import TimestampType
from tecton_core.errors import TectonValidationError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.filter_context import FilterContext
from tecton_core.id_helper import IdHelper
from tecton_core.iterators import batched_iterator
from tecton_core.materialization_context import BoundMaterializationContext
from tecton_core.query.errors import UserCodeError
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.pandas import pipeline_helper
from tecton_core.query.pandas.node import PandasExecNode
from tecton_core.query.pandas.node import SqlExecNode
from tecton_core.schema import Schema
from tecton_core.schema_validation import CastError
from tecton_core.schema_validation import DiffResult
from tecton_core.schema_validation import cast
from tecton_core.schema_validation import cast_columns
from tecton_core.schema_validation import tecton_schema_to_arrow_schema
from tecton_proto.args.pipeline__client_pb2 import DataSourceNode
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.args.pipeline__client_pb2 import TransformationNode
from tecton_proto.args.transformation__client_pb2 import TransformationMode
from tecton_proto.common.data_source_type__client_pb2 import DataSourceType


logger = logging.getLogger(__name__)

# Maps a tecton datatype to the correct pandas datatype which is to be used when an output schema is defined by the user
PRIMITIVE_TECTON_DATA_TYPE_TO_PANDAS_DATA_TYPE = {
    Int32Type(): "int32",
    Int64Type(): "int64",
    Float32Type(): "float32",
    Float64Type(): "float64",
    StringType(): "string",
    BoolType(): "bool",
    TimestampType(): "datetime64[ns]",
}


@attrs.frozen
class PandasDataSourceScanNode(PandasExecNode):
    ds: specs.DataSourceSpec
    ds_node: Optional[DataSourceNode]
    is_stream: bool = attrs.field()
    start_time: Optional[datetime]
    end_time: Optional[datetime]

    def _to_dataframe(self) -> pandas.DataFrame:
        batch_source = self.ds.batch_source
        assert isinstance(batch_source, specs.PandasBatchSourceSpec)

        df = self._get_ds_from_dsf(batch_source, self.ds.name)

        if self.ds.type == DataSourceType.STREAM_WITH_BATCH:
            schema = self.ds.stream_source.spark_schema
            cols = [field.name for field in schema.fields]
            df = df[cols]
        elif self.ds.type == DataSourceType.PUSH_WITH_BATCH:
            schema = self.ds.schema.tecton_schema
            cols = [field.name for field in schema.columns]
            df = df[cols]
        return df

    def _get_ds_from_dsf(self, batch_source: specs.PandasBatchSourceSpec, data_source_name: str) -> pandas.DataFrame:
        function_args = {}
        if batch_source.supports_time_filtering:
            function_args["filter_context"] = FilterContext(self.start_time, self.end_time)
        if batch_source.secrets:
            if self.secret_resolver is None:
                msg = "No secret resolver was provided. For Local Development, please set secrets to string literals."
                raise TectonValidationError(msg)
            function_args["secrets"] = self.secret_resolver.resolve_map(batch_source.secrets)
        try:
            ret = batch_source.function(**function_args)
        except Exception as exc:
            msg = f"Evaluating Pandas data source '{data_source_name}' failed with exception."
            raise UserCodeError(msg) from exc

        if not isinstance(ret, pandas.DataFrame):
            msg = (
                f"The function of Pandas data source '{data_source_name}' is expected to return result "
                f"with Pandas DataFrame type, but returned result with type {type(ret)} instead."
            )
            raise TectonValidationError(msg, can_drop_traceback=True)

        return ret


@attrs.frozen
class PandasFeatureViewPipelineNode(PandasExecNode):
    inputs_map: Dict[str, NodeRef]
    feature_definition_wrapper: FeatureDefinitionWrapper
    feature_time_limits: Optional[pendulum.Period]

    # `check_view_schema` is not actually used in pandas node. Having it here to keep all QT node can be intialized consistently.
    check_view_schema: bool

    def to_arrow(self) -> pyarrow.Table:
        table = self._node_to_value(self.feature_definition_wrapper.pipeline.root)
        return table

    def _to_dataframe(self):
        return self.to_arrow().read_pandas()

    def _node_to_value(
        self, pipeline_node: PipelineNode
    ) -> Union[pyarrow.RecordBatchReader, BoundMaterializationContext]:
        if pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_dataframe(pipeline_node.transformation_node)
        elif pipeline_node.HasField("data_source_node"):
            ds_query_node = self.inputs_map[pipeline_node.data_source_node.input_name].node
            assert isinstance(
                ds_query_node, PandasDataNode
            ), "A PandasFeatureViewPipelineNode cannot operate on standard DataSourceScanNodes. They must have been replaced by PandasDataNodes."
            return ds_query_node.to_arrow()
        elif pipeline_node.HasField("materialization_context_node") or pipeline_node.HasField("context_node"):
            return BoundMaterializationContext.create(self.feature_time_limits.start, self.feature_time_limits.end)
        else:
            msg = f"Pipeline node of kind {pipeline_node.WhichOneof('node_type')} is not supported in Pandas mode"
            raise NotImplementedError(msg)

    def _node_inputs_to_pandas_udf_args(self, node: TransformationNode) -> Tuple[List[Any], Dict[str, Any]]:
        args = []
        kwargs = {}

        for transformation_input in node.inputs:
            node_value = self._node_to_value(transformation_input.node)
            if isinstance(node_value, pyarrow.RecordBatchReader):
                node_value = arrow_to_pandas_dataframe(node_value)
            else:
                assert isinstance(node_value, (BoundMaterializationContext,))

            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                msg = f"Unknown argument type for Input node: {transformation_input}"
                raise KeyError(msg)

        return args, kwargs

    def _node_inputs_to_python_udf_args(self, node: TransformationNode) -> Iterable[Tuple[List[Any], Dict[str, Any]]]:
        """Returns iterator producing tuples (args, kwargs).
        Each tuple represents individual call of python UDF with a single row from input arrow dataframe.
        """
        args = []
        kwargs = {}
        for transformation_input in node.inputs:
            node_value = self._node_to_value(transformation_input.node)
            if isinstance(node_value, pyarrow.RecordBatchReader):

                def row_iterator(reader):
                    while True:
                        try:
                            yield from reader.read_next_batch().to_pylist()
                        except StopIteration:
                            return

                node_value = row_iterator(node_value)
            else:
                assert isinstance(node_value, (BoundMaterializationContext,))
                node_value = itertools.repeat(node_value)

            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                msg = f"Unknown argument type for Input node: {transformation_input}"
                raise KeyError(msg)

        assert args or kwargs
        return zip(
            zip(*args) if args else itertools.repeat([]),
            (dict(zip(kwargs.keys(), kwargs_values)) for kwargs_values in zip(*kwargs.values()))
            if kwargs
            else itertools.repeat({}),
        )

    def _transformation_node_to_dataframe(self, transformation_node: TransformationNode) -> pyarrow.RecordBatchReader:
        """Recursively translates inputs to values and then passes them to the transformation."""

        id_to_transformation = {t.id: t for t in self.feature_definition_wrapper.transformations}
        transformation = id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
        user_function = transformation.user_function

        if transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
            return self._call_pandas_udf_and_verify_result(
                transformation_node, user_function, self.feature_definition_wrapper.view_schema
            )

        if transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYTHON:
            return self._call_python_udf_and_verify_result(
                transformation_node, user_function, self.feature_definition_wrapper.view_schema
            )

        msg = (
            f"Unknown transformation mode {transformation.transformation_mode} "
            f"in feature view {self.feature_definition_wrapper.fv_spec.name}"
        )
        raise TectonValidationError(msg)

    def _call_pandas_udf_and_verify_result(
        self,
        transformation_node: TransformationNode,
        user_function: typing.Callable,
        expected_schema: Schema,
    ) -> pyarrow.RecordBatchReader:
        args, kwargs = self._node_inputs_to_pandas_udf_args(transformation_node)
        try:
            ret = user_function(*args, **kwargs)
        except Exception as exc:
            msg = (
                "Pandas pipeline function (feature view "
                f"'{self.feature_definition_wrapper.fv_spec.name}') "
                f"failed with exception"
            )
            raise UserCodeError(msg) from exc
        try:
            if not isinstance(ret, pandas.DataFrame):
                msg = f"Expected a Pandas Dataframe, but returned {type(ret)}"
                raise CastError(msg)
            return cast(ret, schema=expected_schema).to_reader()
        except CastError as exc:
            validation_msg = (
                f"Pipeline function (feature view '{self.feature_definition_wrapper.fv_spec.name}') produced an "
                + "invalid result: "
            )
            raise CastError(validation_msg + str(exc)) from None

    def _call_python_udf_and_verify_result(
        self,
        transformation_node: TransformationNode,
        user_function: typing.Callable,
        expected_schema: Schema,
    ) -> pyarrow.RecordBatchReader:
        input_it = self._node_inputs_to_python_udf_args(transformation_node)

        def record_batch_iterator():
            for input_batch in batched_iterator(input_it, batch_size=1_000_000):
                columns = {name: [] for name in expected_schema.column_names()}
                try:
                    for args, kwargs in input_batch:
                        try:
                            ret = user_function(*args, **kwargs)
                            if not isinstance(ret, dict):
                                other_type = type(ret)
                                msg = f"Expected a Python dict, but returned {other_type}"
                                raise CastError(msg)
                        except Exception as exc:
                            msg = (
                                "Python pipeline function (feature view "
                                f"'{self.feature_definition_wrapper.fv_spec.name}') "
                                f"failed with exception"
                            )
                            raise UserCodeError(msg) from exc
                        for name in columns:
                            if name not in ret:
                                raise CastError.for_diff(DiffResult(missing_fields=[name]))
                            columns[name].append(ret[name])

                    def column_array(name: str, dtype: pyarrow.DataType) -> pyarrow.Array:
                        col = columns[name]
                        return pyarrow.array(col, type=dtype, size=len(col))

                    arrow_schema = tecton_schema_to_arrow_schema(expected_schema)
                    arrays = cast_columns(
                        column_getter=column_array, schema=tecton_schema_to_arrow_schema(expected_schema)
                    )
                    yield pyarrow.RecordBatch.from_arrays(arrays, schema=arrow_schema)
                except CastError as e:
                    validation_msg = (
                        f"Pipeline function (feature view '{self.feature_definition_wrapper.fv_spec.name}') produced an "
                        + "invalid result: "
                    )
                    raise CastError(validation_msg + str(e)) from None

        schema = tecton_schema_to_arrow_schema(expected_schema)
        return pyarrow.RecordBatchReader.from_batches(schema, record_batch_iterator())


@attrs.frozen
class PandasDataNode(PandasExecNode):
    input_reader: pyarrow.RecordBatchReader

    def to_arrow(self):
        return self.input_reader

    def _to_dataframe(self):
        return self.to_arrow().read_pandas()


@attrs.frozen
class PandasMultiOdfvPipelineNode(PandasExecNode):
    input_node: Union[PandasExecNode, SqlExecNode]
    feature_definition_namespaces: List[Tuple[FeatureDefinitionWrapper, str]]
    use_namespace_feature_prefix: bool
    events_df_timestamp_field: str

    def _to_dataframe(self):
        # ToDo(Oleksii): this method should be deprecated together with Snowflake & Athena computes

        if isinstance(self.input_node, PandasExecNode):
            output_df = arrow_to_pandas_dataframe(self.input_node.to_arrow())
        else:
            output_df = self.input_node._to_dataframe()
        # Apply each ODFV sequentially. Note that attempting to apply ODFV
        # udfs in parallel as we traverse data rows does not meaningfully
        # speed up execution on Athena (unlike in Spark).
        for fdw, namespace in self.feature_definition_namespaces:
            odfv_transformation_node = fdw.pipeline.root.transformation_node
            odfv_inputs = dict(
                self._extract_inputs_for_odfv_from_data(
                    transformation_node=odfv_transformation_node, data_df=output_df, odfv=fdw
                )
            )
            odfv_result_df = self._get_odfv_output_df(odfv_inputs, fdw, namespace, len(output_df))
            output_df = output_df.merge(odfv_result_df, left_index=True, right_index=True)
        return output_df

    def to_arrow(self) -> pyarrow.RecordBatchReader:
        assert isinstance(self.input_node, PandasDataNode)  # only Rift compute is supported
        reader = self.input_node.to_arrow()

        output_schema = reader.schema
        for fdw, namespace in self.feature_definition_namespaces:
            arrow_schema = tecton_schema_to_arrow_schema(fdw.view_schema)
            for name, type_ in zip(arrow_schema.names, arrow_schema.types):
                if self.use_namespace_feature_prefix:
                    mapped_name = self.column_name_updater(f"{namespace}{fdw.namespace_separator}{name}")
                else:
                    mapped_name = name
                output_schema = output_schema.append(pyarrow.field(mapped_name, type_))

        def batch_iter():
            while True:
                try:
                    batch = reader.read_next_batch()
                except StopIteration:
                    return

                output_columns = batch.columns

                for fdw, namespace in self.feature_definition_namespaces:
                    odfv_transformation_node = fdw.pipeline.root.transformation_node
                    odfv_inputs = dict(
                        self._extract_inputs_for_odfv_from_data(
                            transformation_node=odfv_transformation_node,
                            data_df=batch,
                            odfv=fdw,
                        )
                    )
                    odfv_result_df = self._run_odfv(odfv_inputs, fdw, batch.num_rows)
                    try:
                        odfv_result = cast(odfv_result_df, fdw.view_schema)
                    except CastError as exc:
                        msg = f"ODFV '{fdw.name}' produced unexpected result that doesn't match output schema: "
                        raise CastError(msg + str(exc)) from None

                    for column in odfv_result.columns:
                        output_columns.append(column.combine_chunks())

                yield pyarrow.RecordBatch.from_arrays(output_columns, schema=output_schema)

        return pyarrow.RecordBatchReader.from_batches(output_schema, batch_iter())

    def _get_odfv_output_df(
        self,
        input_df: pandas.DataFrame,
        fdw: FeatureDefinitionWrapper,
        namespace: str,
        num_rows: int,
    ) -> pandas.DataFrame:
        odfv_result_df = self._run_odfv(input_df, fdw, num_rows)
        rename_map = {}
        datatypes = {}
        # Namespace ODFV outputs to this FV to avoid conflicts in output schemas
        # with other FV
        output_schema = fdw.view_schema.column_name_and_data_types()
        for column_name, datatype in output_schema:
            if self.use_namespace_feature_prefix:
                mapped_name = self.column_name_updater(f"{namespace}{fdw.namespace_separator}{column_name}")
            else:
                mapped_name = column_name
            rename_map[column_name] = mapped_name
            if datatype in PRIMITIVE_TECTON_DATA_TYPE_TO_PANDAS_DATA_TYPE:
                datatypes[mapped_name] = PRIMITIVE_TECTON_DATA_TYPE_TO_PANDAS_DATA_TYPE[datatype]
        return odfv_result_df.rename(columns=rename_map)[[*rename_map.values()]].astype(datatypes)

    def _run_odfv(
        self, odfv_inputs: pandas.DataFrame, odfv: FeatureDefinitionWrapper, num_rows: int
    ) -> pandas.DataFrame:
        odfv_pipeline = pipeline_helper.build_odfv_execution_pipeline(
            pipeline=odfv.pipeline, transformations=odfv.transformations, name=odfv.name
        )

        if odfv.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
            odfv_result_df = odfv_pipeline.execute_with_inputs(odfv_inputs)
            return odfv_result_df
        elif odfv.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYTHON:
            # The inputs are currently a mapping of input_name to pandas DF
            # We need turn the ODFV inputs from a pandas DF to a list of dictionaries
            # Then we need to iterate through all rows of the input data set, pass the input dicts into the ODFV
            # And finally convert the resulting list of dicts into a pandas DF
            for input_name in odfv_inputs.keys():
                # Map pandas DFs to List of dicts (one dict per row)
                odfv_inputs[input_name] = odfv_inputs[input_name].to_dict("records")

            odfv_result_list = []

            for row_index in range(num_rows):
                # Iterate through all rows of the data and invoke the ODFV
                row_odfv_inputs = {}
                for input_name in odfv_inputs.keys():
                    row_odfv_inputs[input_name] = odfv_inputs[input_name][row_index]

                odfv_result_dict = odfv_pipeline.execute_with_inputs(row_odfv_inputs)
                odfv_result_list.append(odfv_result_dict)
            return pandas.DataFrame.from_dict(odfv_result_list)
        else:
            msg = f"ODFV {odfv.name} has an unexpected transformation mode: {odfv.transformation_mode}"
            raise TectonValidationError(msg)

    def _extract_inputs_for_odfv_from_data(
        self,
        transformation_node: TransformationNode,
        data_df: Union[pandas.DataFrame, pyarrow.RecordBatch],
        odfv: FeatureDefinitionWrapper,
    ) -> Iterable[Tuple[str, pandas.DataFrame]]:
        for input in transformation_node.inputs:
            if input.node.HasField("request_data_source_node"):
                input_name = input.node.request_data_source_node.input_name
                request_context_schema = input.node.request_data_source_node.request_context.tecton_schema

                if isinstance(data_df, pandas.DataFrame):
                    # ToDo(Oleksii): remove this path once Snowflake & Athena are deprecated
                    request_context_fields = [self.column_name_updater(c.name) for c in request_context_schema.columns]
                    for f in request_context_fields:
                        if f not in data_df.columns:
                            msg = f"ODFV {odfv.name} has a dependency on the Request Data Source named '{input_name}'. Field {f} of this Request Data Source is not found in the spine. Available columns: {list(data_df.columns)}"
                            raise TectonValidationError(msg)

                    input_df = data_df[request_context_fields]

                elif isinstance(data_df, pyarrow.RecordBatch):
                    try:
                        input_df = cast(data_df, Schema(request_context_schema)).to_pandas()
                    except CastError as exc:
                        msg = f"ODFV {odfv.name} has a dependency on the Request Data Source named '{input_name}', but it didn't pass schema validation: "
                        raise CastError(msg + str(exc)) from None
                else:
                    msg = f"Unexpected input dataframe type: {type(data_df)}"
                    raise RuntimeError(msg)

                yield input_name, input_df
            elif input.node.HasField("feature_view_node"):
                input_name = input.node.feature_view_node.input_name
                fv_features = feature_set_config.find_dependent_feature_set_items(
                    odfv.fco_container, input.node, {}, odfv.id
                )[0]
                # Generate dependent column mappings since dependent FV have
                # internal column names with _udf_internal
                select_columns_and_rename_map = {}
                for f in fv_features.features:
                    column_name = self.column_name_updater(f"{fv_features.namespace}__{f}")
                    mapped_name = self.column_name_updater(f)
                    select_columns_and_rename_map[column_name] = mapped_name
                if isinstance(data_df, pandas.DataFrame):
                    # ToDo(Oleksii): remove this path once Snowflake & Athena are deprecated
                    for f in select_columns_and_rename_map.keys():
                        if f not in data_df.columns:
                            msg = f"ODFV {odfv.name} has a dependency on the Feature View '{input_name}'. Feature {f} of this Feature View is not found in the retrieved historical data. Available columns: {list(data_df.columns)}"
                            raise TectonValidationError(msg)
                    # Let's select all of the features of the input FV from data_df
                    input_df = data_df.rename(columns=select_columns_and_rename_map)[
                        [*select_columns_and_rename_map.values()]
                    ]
                elif isinstance(data_df, pyarrow.RecordBatch):
                    columns = []
                    for f in select_columns_and_rename_map:
                        try:
                            columns.append(data_df.column(f))
                        except KeyError:
                            msg = f"ODFV {odfv.name} has a dependency on the Feature View '{input_name}'. Feature {f} of this Feature View is not found in the retrieved historical data. Available columns: {list(data_df.column_names)}"
                            raise TectonValidationError(msg)

                    input_df = pyarrow.RecordBatch.from_arrays(
                        columns, names=list(select_columns_and_rename_map.values())
                    ).to_pandas()
                else:
                    msg = f"Unexpected input dataframe type: {type(data_df)}"
                    raise RuntimeError(msg)

                yield input_name, input_df
            elif input.node.HasField("transformation_node"):
                yield from self._extract_inputs_for_odfv_from_data(
                    transformation_node=input.node.transformation_node,
                    data_df=data_df,
                    odfv=odfv,
                )
            else:
                msg = f"Unexpected input found ({input.arg_name}) on ODFV {odfv.name}"
                raise Exception(msg)


@attrs.frozen
class PandasRenameColsNode(PandasExecNode):
    input_node: Union[PandasExecNode, SqlExecNode]
    mapping: Optional[Dict[str, str]]
    drop: Optional[List[str]]

    def _to_dataframe(self) -> pandas.DataFrame:
        input_df = self.input_node._to_dataframe()
        output_df = input_df
        if self.drop:
            columns_to_drop = [self.column_name_updater(name) for name in self.drop]
            output_df = input_df.drop(columns=columns_to_drop)
        if self.mapping:
            output_df = input_df.rename(self.mapping)
        return output_df

    def to_arrow(self) -> pyarrow.RecordBatchReader:
        reader = self.input_node.to_arrow()

        input_schema = reader.schema
        output_fields = []
        for name in input_schema.names:
            if self.drop and name in self.drop:
                continue

            field = input_schema.field(name)
            if self.mapping and name in self.mapping:
                field = field.with_name(self.mapping[name])

            output_fields.append(field)

        output_schema = pyarrow.schema(output_fields)

        def batch_iter():
            while True:
                try:
                    batch = reader.read_next_batch()
                except StopIteration:
                    return

                output_columns = []
                for name, column in zip(batch.column_names, batch.columns):
                    if self.drop and name in self.drop:
                        continue

                    output_columns.append(column)

                yield pyarrow.RecordBatch.from_arrays(output_columns, schema=output_schema)

        return pyarrow.RecordBatchReader.from_batches(output_schema, batch_iter())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auditlog/metadata__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,tecton_proto/auditlog/metadata__client.proto\x12\x15tecton_proto.auditlog\x1a google/protobuf/descriptor.proto\"\x96\x01\n\x10\x41uditLogMetadata\x12+\n\x12write_to_audit_log\x18\x01 \x01(\x08R\x0fwriteToAuditLog\x12;\n\x1a\x63ustomer_facing_event_name\x18\x02 \x01(\tR\x17\x63ustomerFacingEventName\x12\x18\n\x07version\x18\x03 \x01(\tR\x07version\"\x9c\x01\n\x0f\x41uditLogOptions\x12\x41\n\nvisibility\x18\x01 \x01(\x0e\x32!.tecton_proto.auditlog.VisibilityR\nvisibility\x12\x46\n\ttransform\x18\x02 \x01(\x0e\x32(.tecton_proto.auditlog.AuditLogTransformR\ttransform*U\n\nVisibility\x12\x1a\n\x16VISIBILITY_UNSPECIFIED\x10\x00\x12\x13\n\x0fVISIBILITY_OMIT\x10\x01\x12\x16\n\x12VISIBILITY_VISIBLE\x10\x02*g\n\x11\x41uditLogTransform\x12\x19\n\x15TRANSFORM_UNSPECIFIED\x10\x00\x12\x1a\n\x16TRANSFORM_LEGACY_ROLES\x10\x01\x12\x1b\n\x17TRANSFORM_REDACT_STRING\x10\x02:b\n\x07options\x12\x1d.google.protobuf.FieldOptions\x18\x9d\xd3\xe7\x08 \x01(\x0b\x32&.tecton_proto.auditlog.AuditLogOptionsR\x07options:x\n\x12\x61udit_log_metadata\x12\x1e.google.protobuf.MethodOptions\x18\x9e\xd3\xe7\x08 \x01(\x0b\x32\'.tecton_proto.auditlog.AuditLogMetadataR\x10\x61uditLogMetadataBC\n\x13\x63om.tecton.auditlogP\x01Z*github.com/tecton-ai/tecton_proto/auditlog')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auditlog.metadata__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
  google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(options)
  google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(audit_log_metadata)

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\023com.tecton.auditlogP\001Z*github.com/tecton-ai/tecton_proto/auditlog'
  _VISIBILITY._serialized_start=417
  _VISIBILITY._serialized_end=502
  _AUDITLOGTRANSFORM._serialized_start=504
  _AUDITLOGTRANSFORM._serialized_end=607
  _AUDITLOGMETADATA._serialized_start=106
  _AUDITLOGMETADATA._serialized_end=256
  _AUDITLOGOPTIONS._serialized_start=259
  _AUDITLOGOPTIONS._serialized_end=415
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/remoteenvironmentservice/remote_environment_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.auth import principal__client_pb2 as tecton__proto_dot_auth_dot_principal____client__pb2
from tecton_proto.auth import service__client_pb2 as tecton__proto_dot_auth_dot_service____client__pb2
from tecton_proto.common import container_image__client_pb2 as tecton__proto_dot_common_dot_container__image____client__pb2
from tecton_proto.data import remote_compute_environment__client_pb2 as tecton__proto_dot_data_dot_remote__compute__environment____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nNtecton_proto/remoteenvironmentservice/remote_environment_service__client.proto\x12%tecton_proto.remoteenvironmentservice\x1a)tecton_proto/auth/principal__client.proto\x1a\'tecton_proto/auth/service__client.proto\x1a\x31tecton_proto/common/container_image__client.proto\x1a:tecton_proto/data/remote_compute_environment__client.proto\"\xb5\x04\n\x1e\x43reateRemoteEnvironmentRequest\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\x12\x42\n\nimage_info\x18\x03 \x01(\x0b\x32#.tecton_proto.common.ContainerImageR\timageInfo\x12\x0e\n\x02id\x18\x04 \x01(\tR\x02id\x12%\n\x0epython_version\x18\x05 \x01(\tR\rpythonVersion\x12\"\n\x0crequirements\x18\x06 \x01(\tR\x0crequirements\x12\x33\n\x15resolved_requirements\x18\x07 \x01(\tR\x14resolvedRequirements\x12,\n\x12s3_wheels_location\x18\x0e \x01(\tR\x10s3WheelsLocation\x12:\n\x19transform_runtime_version\x18\x0f \x01(\tR\x17transformRuntimeVersion\x12O\n$rift_materialization_runtime_version\x18\x10 \x01(\tR!riftMaterializationRuntimeVersion\x12\x1f\n\x0bsdk_version\x18\x11 \x01(\tR\nsdkVersion\x12-\n\x12online_provisioned\x18\x12 \x01(\x08R\x11onlineProvisioned\"}\n\x1f\x43reateRemoteEnvironmentResponse\x12Z\n\x12remote_environment\x18\x01 \x01(\x0b\x32+.tecton_proto.data.RemoteComputeEnvironmentR\x11remoteEnvironment\"\xac\x01\n\x1eUpdateRemoteEnvironmentRequest\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x36\n\x17remote_function_version\x18\x02 \x01(\tR\x15remoteFunctionVersion\x12\x42\n\x06status\x18\x03 \x01(\x0e\x32*.tecton_proto.data.RemoteEnvironmentStatusR\x06status\"}\n\x1fUpdateRemoteEnvironmentResponse\x12Z\n\x12remote_environment\x18\x01 \x01(\x0b\x32+.tecton_proto.data.RemoteComputeEnvironmentR\x11remoteEnvironment\"3\n\x1f\x44\x65leteRemoteEnvironmentsRequest\x12\x10\n\x03ids\x18\x01 \x03(\tR\x03ids\"\"\n DeleteRemoteEnvironmentsResponse\"\x1f\n\x1dListRemoteEnvironmentsRequest\"~\n\x1eListRemoteEnvironmentsResponse\x12\\\n\x13remote_environments\x18\x01 \x03(\x0b\x32+.tecton_proto.data.RemoteComputeEnvironmentR\x12remoteEnvironments\"-\n\x1bGetRemoteEnvironmentRequest\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\"z\n\x1cGetRemoteEnvironmentResponse\x12Z\n\x12remote_environment\x18\x01 \x01(\x0b\x32+.tecton_proto.data.RemoteComputeEnvironmentR\x11remoteEnvironment\"\x8f\x01\n\x1bGetPackagesUploadUrlRequest\x12%\n\x0e\x65nvironment_id\x18\x01 \x01(\tR\renvironmentId\x12I\n\x0bupload_part\x18\x02 \x01(\x0b\x32(.tecton_proto.data.ObjectStoreUploadPartR\nuploadPart\"=\n\x1cGetPackagesUploadUrlResponse\x12\x1d\n\nupload_url\x18\x01 \x01(\tR\tuploadUrl\"C\n\x1aStartPackagesUploadRequest\x12%\n\x0e\x65nvironment_id\x18\x01 \x01(\tR\renvironmentId\"n\n\x1bStartPackagesUploadResponse\x12O\n\x0bupload_info\x18\x01 \x01(\x0b\x32..tecton_proto.data.RemoteEnvironmentUploadInfoR\nuploadInfo\"p\n\x1d\x43ompletePackagesUploadRequest\x12O\n\x0bupload_info\x18\x01 \x01(\x0b\x32..tecton_proto.data.RemoteEnvironmentUploadInfoR\nuploadInfo\"K\n\x1e\x43ompletePackagesUploadResponse\x12)\n\x10storage_location\x18\x01 \x01(\tR\x0fstorageLocation\"K\n\"GetDependentFeatureServicesRequest\x12%\n\x0e\x65nvironment_id\x18\x01 \x01(\tR\renvironmentId\"\x8f\x01\n#GetDependentFeatureServicesResponse\x12h\n\x1a\x64\x65pendent_feature_services\x18\x01 \x03(\x0b\x32*.tecton_proto.data.DependentFeatureServiceR\x18\x64\x65pendentFeatureServices2\xb8\x0f\n\x18RemoteEnvironmentService\x12\xda\x01\n\x17\x43reateRemoteEnvironment\x12\x45.tecton_proto.remoteenvironmentservice.CreateRemoteEnvironmentRequest\x1a\x46.tecton_proto.remoteenvironmentservice.CreateRemoteEnvironmentResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xda\x01\n\x17UpdateRemoteEnvironment\x12\x45.tecton_proto.remoteenvironmentservice.UpdateRemoteEnvironmentRequest\x1a\x46.tecton_proto.remoteenvironmentservice.UpdateRemoteEnvironmentResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xdd\x01\n\x18\x44\x65leteRemoteEnvironments\x12\x46.tecton_proto.remoteenvironmentservice.DeleteRemoteEnvironmentsRequest\x1aG.tecton_proto.remoteenvironmentservice.DeleteRemoteEnvironmentsResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xd4\x01\n\x16ListRemoteEnvironments\x12\x44.tecton_proto.remoteenvironmentservice.ListRemoteEnvironmentsRequest\x1a\x45.tecton_proto.remoteenvironmentservice.ListRemoteEnvironmentsResponse\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organization\x12\xce\x01\n\x14GetRemoteEnvironment\x12\x42.tecton_proto.remoteenvironmentservice.GetRemoteEnvironmentRequest\x1a\x43.tecton_proto.remoteenvironmentservice.GetRemoteEnvironmentResponse\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organization\x12\xce\x01\n\x14GetPackagesUploadUrl\x12\x42.tecton_proto.remoteenvironmentservice.GetPackagesUploadUrlRequest\x1a\x43.tecton_proto.remoteenvironmentservice.GetPackagesUploadUrlResponse\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organization\x12\xcb\x01\n\x13StartPackagesUpload\x12\x41.tecton_proto.remoteenvironmentservice.StartPackagesUploadRequest\x1a\x42.tecton_proto.remoteenvironmentservice.StartPackagesUploadResponse\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organization\x12\xd4\x01\n\x16\x43ompletePackagesUpload\x12\x44.tecton_proto.remoteenvironmentservice.CompletePackagesUploadRequest\x1a\x45.tecton_proto.remoteenvironmentservice.CompletePackagesUploadResponse\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organization\x12\xe3\x01\n\x1bGetDependentFeatureServices\x12I.tecton_proto.remoteenvironmentservice.GetDependentFeatureServicesRequest\x1aJ.tecton_proto.remoteenvironmentservice.GetDependentFeatureServicesResponse\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organizationB\x82\x01\n#com.tecton.remoteenvironmentserviceB\x1dRemoteEnvironmentServiceProtoP\x01Z:github.com/tecton-ai/tecton_proto/remoteenvironmentservice')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.remoteenvironmentservice.remote_environment_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n#com.tecton.remoteenvironmentserviceB\035RemoteEnvironmentServiceProtoP\001Z:github.com/tecton-ai/tecton_proto/remoteenvironmentservice'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['CreateRemoteEnvironment']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['CreateRemoteEnvironment']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['UpdateRemoteEnvironment']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['UpdateRemoteEnvironment']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['DeleteRemoteEnvironments']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['DeleteRemoteEnvironments']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['ListRemoteEnvironments']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['ListRemoteEnvironments']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['GetRemoteEnvironment']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['GetRemoteEnvironment']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['GetPackagesUploadUrl']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['GetPackagesUploadUrl']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['StartPackagesUpload']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['StartPackagesUpload']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['CompletePackagesUpload']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['CompletePackagesUpload']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _REMOTEENVIRONMENTSERVICE.methods_by_name['GetDependentFeatureServices']._options = None
  _REMOTEENVIRONMENTSERVICE.methods_by_name['GetDependentFeatureServices']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _CREATEREMOTEENVIRONMENTREQUEST._serialized_start=317
  _CREATEREMOTEENVIRONMENTREQUEST._serialized_end=882
  _CREATEREMOTEENVIRONMENTRESPONSE._serialized_start=884
  _CREATEREMOTEENVIRONMENTRESPONSE._serialized_end=1009
  _UPDATEREMOTEENVIRONMENTREQUEST._serialized_start=1012
  _UPDATEREMOTEENVIRONMENTREQUEST._serialized_end=1184
  _UPDATEREMOTEENVIRONMENTRESPONSE._serialized_start=1186
  _UPDATEREMOTEENVIRONMENTRESPONSE._serialized_end=1311
  _DELETEREMOTEENVIRONMENTSREQUEST._serialized_start=1313
  _DELETEREMOTEENVIRONMENTSREQUEST._serialized_end=1364
  _DELETEREMOTEENVIRONMENTSRESPONSE._serialized_start=1366
  _DELETEREMOTEENVIRONMENTSRESPONSE._serialized_end=1400
  _LISTREMOTEENVIRONMENTSREQUEST._serialized_start=1402
  _LISTREMOTEENVIRONMENTSREQUEST._serialized_end=1433
  _LISTREMOTEENVIRONMENTSRESPONSE._serialized_start=1435
  _LISTREMOTEENVIRONMENTSRESPONSE._serialized_end=1561
  _GETREMOTEENVIRONMENTREQUEST._serialized_start=1563
  _GETREMOTEENVIRONMENTREQUEST._serialized_end=1608
  _GETREMOTEENVIRONMENTRESPONSE._serialized_start=1610
  _GETREMOTEENVIRONMENTRESPONSE._serialized_end=1732
  _GETPACKAGESUPLOADURLREQUEST._serialized_start=1735
  _GETPACKAGESUPLOADURLREQUEST._serialized_end=1878
  _GETPACKAGESUPLOADURLRESPONSE._serialized_start=1880
  _GETPACKAGESUPLOADURLRESPONSE._serialized_end=1941
  _STARTPACKAGESUPLOADREQUEST._serialized_start=1943
  _STARTPACKAGESUPLOADREQUEST._serialized_end=2010
  _STARTPACKAGESUPLOADRESPONSE._serialized_start=2012
  _STARTPACKAGESUPLOADRESPONSE._serialized_end=2122
  _COMPLETEPACKAGESUPLOADREQUEST._serialized_start=2124
  _COMPLETEPACKAGESUPLOADREQUEST._serialized_end=2236
  _COMPLETEPACKAGESUPLOADRESPONSE._serialized_start=2238
  _COMPLETEPACKAGESUPLOADRESPONSE._serialized_end=2313
  _GETDEPENDENTFEATURESERVICESREQUEST._serialized_start=2315
  _GETDEPENDENTFEATURESERVICESREQUEST._serialized_end=2390
  _GETDEPENDENTFEATURESERVICESRESPONSE._serialized_start=2393
  _GETDEPENDENTFEATURESERVICESRESPONSE._serialized_end=2536
  _REMOTEENVIRONMENTSERVICE._serialized_start=2539
  _REMOTEENVIRONMENTSERVICE._serialized_end=4515
# @@protoc_insertion_point(module_scope)



from typing import Optional


_trace_id = None


def set_trace_id(trace_id: Optional[str]) -> None:
    global _trace_id
    _trace_id = trace_id


def get_trace_id() -> Optional[str]:
    return _trace_id



from __future__ import annotations

from enum import Enum
from typing import List

import attrs

from tecton_core import schema


class TextEmbeddingModel(str, Enum):
    ALL_MINILM_L6_V2 = "sentence-transformers/all-MiniLM-L6-v2"
    MXBAI_EMBED_LARGE_v1 = "mixedbread-ai/mxbai-embed-large-v1"
    BGE_LARGE_EN_v1_5 = "BAAI/bge-large-en-v1.5"
    BGE_BASE_EN_v1_5 = "BAAI/bge-base-en-v1.5"
    BGE_SMALL_EN_v1_5 = "BAAI/bge-small-en-v1.5"
    GTE_LARGE = "thenlper/gte-large"
    GTE_BASE = "thenlper/gte-base"
    GTE_GTE_SMALL = "thenlper/gte-small"
    SNOWFLAKE_ARCTIC_EMBED_XS = "Snowflake/snowflake-arctic-embed-xs"
    SNOWFLAKE_ARCTIC_EMBED_S = "Snowflake/snowflake-arctic-embed-s"
    SNOWFLAKE_ARCTIC_EMBED_M = "Snowflake/snowflake-arctic-embed-m"
    SNOWFLAKE_ARCTIC_EMBED_L = "Snowflake/snowflake-arctic-embed-l"
    E5_LARGE_UNSUPERVISED = "intfloat/e5-large-unsupervised"
    E5_BASE_UNSUPERVISED = "intfloat/e5-base-unsupervised"
    E5_SMALL_UNSUPERVISED = "intfloat/e5-small-unsupervised"


class BaseInferenceConfig:
    pass


@attrs.frozen
class TextEmbeddingInferenceConfig(BaseInferenceConfig):
    input_column: str
    output_column: str
    model: TextEmbeddingModel


# TODO(jiadong): Consolidate TextEmbeddingInferenceConfig and this config and migrate pre-trained models to use it.
@attrs.frozen
class CustomModelConfig(BaseInferenceConfig):
    input_columns: List[schema.Column]
    output_column: schema.Column
    model_data_dir: str



from typing import Dict

import attrs

from tecton_core import schema
from tecton_core.data_types import DataType
from tecton_proto.args import pipeline__client_pb2 as pipeline_pb2


@attrs.frozen
class RequestContext:
    """
    Wrapper around the RequestContext proto object (that is part of the RequestDataSourceNode in the FV pipeline).
    """

    schema: Dict[str, DataType]

    @classmethod
    def from_proto(cls, proto: pipeline_pb2.RequestContext) -> "RequestContext":
        rc_schema = schema.Schema(proto.tecton_schema)
        return RequestContext(schema=rc_schema.to_dict())

    def merge(self, other):
        for field in other.schema:
            if field in self.schema:
                assert self.schema[field] == other.schema[field], f"Mismatched request context field types for {field}"
            else:
                self.schema[field] = other.schema[field]



import logging
import tempfile
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas
import pendulum
from pyspark.sql import DataFrame
from pyspark.sql.streaming import StreamingQuery

import tecton
from tecton._internals import errors
from tecton._internals import mock_source_utils
from tecton._internals.rewrite import MockDataRewrite
from tecton.framework.data_frame import TectonDataFrame
from tecton.run_api_consts import AGGREGATION_LEVEL_DISABLED
from tecton.run_api_consts import AGGREGATION_LEVEL_FULL
from tecton.run_api_consts import AGGREGATION_LEVEL_PARTIAL
from tecton.run_api_consts import DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME
from tecton.run_api_consts import DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME
from tecton.run_api_consts import SUPPORTED_AGGREGATION_LEVEL_VALUES
from tecton.tecton_context import TectonContext
from tecton_core import pipeline_common
from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.errors import TectonValidationError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.pipeline_common import get_all_input_keys
from tecton_core.pipeline_common import get_fco_ids_to_input_keys
from tecton_core.query.builder import build_aggregated_time_range_run_query
from tecton_core.query.builder import build_materialization_querytree
from tecton_core.query.builder import build_pipeline_querytree
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import AddAnchorTimeNode
from tecton_core.query.nodes import ConvertEpochToTimestampNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query_consts import anchor_time
from tecton_core.query_consts import window_end_column_name
from tecton_core.realtime_context import RealtimeContext
from tecton_core.schema import Schema
from tecton_proto.args import pipeline__client_pb2 as pipeline_pb2
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.transformation__client_pb2 import TransformationMode
from tecton_spark import materialization_plan
from tecton_spark.partial_aggregations import partial_aggregate_column_renames
from tecton_spark.rtfv_pipeline import RealtimeFeaturePipeline
from tecton_spark.spark_helper import check_spark_version


logger = logging.getLogger(__name__)


def maybe_warn_incorrect_time_range_size(
    fd: FeatureDefinitionWrapper, start_time: datetime, end_time: datetime, aggregation_level: Optional[str]
):
    time_range = end_time - start_time
    if fd.is_temporal_aggregate:
        if fd.is_continuous:
            # There should not be any time range warnings for continuous aggregates.
            return
        slide_interval = fd.aggregate_slide_interval.ToTimedelta()
        if aggregation_level == AGGREGATION_LEVEL_FULL:
            if not fd.has_lifetime_aggregate:
                max_aggregation_window = -fd.earliest_window_start
                if time_range < max_aggregation_window:
                    logger.warning(
                        f"Run time range ({start_time}, {end_time}) is smaller than the maximum aggregation size: {max_aggregation_window}. This may lead to incorrect aggregate feature values."
                    )

            if time_range.total_seconds() % slide_interval.total_seconds() != 0:
                logger.warning(
                    f"Run time range ({start_time}, {end_time}) is not a multiple of the aggregation_interval: {slide_interval}. This may lead to incorrect aggregate feature values, since Tecton pre-aggregates data in smaller time windows based on the aggregation_interval size."
                )
        elif aggregation_level == AGGREGATION_LEVEL_PARTIAL:
            if time_range.total_seconds() % slide_interval.total_seconds() != 0:
                logger.warning(
                    f"Run time range ({start_time}, {end_time}) is not a multiple of the aggregation_interval: {slide_interval}. This may lead to incorrect aggregate feature values, since Tecton pre-aggregates data in smaller time windows based on the aggregation_interval size."
                )
    elif fd.is_incremental_backfill and time_range != fd.batch_materialization_schedule:
        logger.warning(
            f"Run time range ({start_time}, {end_time}) is not equivalent to the batch_schedule: {fd.batch_materialization_schedule}. This may lead to incorrect feature values since feature views with incremental_backfills typically implicitly rely on the materialization range being equivalent to the batch_schedule."
        )


def run_batch(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fd: FeatureDefinitionWrapper,
    feature_start_time: datetime,
    feature_end_time: datetime,
    mock_data_sources: Dict[str, NodeRef],
    aggregation_level: Optional[str],
) -> "tecton.framework.data_frame.TectonDataFrame":
    _print_run_deprecation_message(aggregation_level)

    if not fd.is_realtime:
        check_spark_version(fd.fv_spec.batch_cluster_config)

    return _querytree_run_batch(
        dialect=dialect,
        compute_mode=compute_mode,
        fd=fd,
        feature_start_time=feature_start_time,
        feature_end_time=feature_end_time,
        mock_data_sources=mock_data_sources,
        aggregation_level=aggregation_level,
    )


def _build_run_batch_querytree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fd: FeatureDefinitionWrapper,
    feature_end_time: datetime,
    feature_time_limits_aligned: pendulum.Period,
    aggregation_level: Optional[str],
) -> NodeRef:
    """Build run_batch query tree

    This assumes that inputs are validated already and is general (should not
    handle mocking data). Using mock data is considered a query tree rewrite.

    Any extra querytree nodes in this function should simply be a display-level
    modification (like field rename, type change, etc).
    """
    if fd.is_temporal:
        qt = build_materialization_querytree(
            dialect, compute_mode, fd, for_stream=False, feature_data_time_limits=feature_time_limits_aligned
        )
        # For a BFV, the materialization querytree is an `AddAnchorTimeNode` wrapped around exactly what we want, so we
        # just extract the input node.
        assert isinstance(qt.node, AddAnchorTimeNode)
        return qt.node.inputs[0]
    elif fd.is_temporal_aggregate:
        if aggregation_level == AGGREGATION_LEVEL_PARTIAL:
            qt = build_materialization_querytree(
                dialect,
                compute_mode,
                fd,
                for_stream=False,
                feature_data_time_limits=feature_time_limits_aligned,
                include_window_end_time=True,
                aggregation_anchor_time=feature_end_time,
            )
            if fd.is_continuous:
                renames = {
                    **partial_aggregate_column_renames(
                        slide_interval_string=fd.get_aggregate_slide_interval_string,
                        trailing_time_window_aggregation=fd.trailing_time_window_aggregation(),
                    ),
                }
                drop = [anchor_time()]
            else:
                # The `PartialAggNode` returned by `build_materialization_querytree` converts timestamps to epochs. We convert back
                # from epochs to timestamps since timestamps are more readable.
                qt = ConvertEpochToTimestampNode(
                    dialect,
                    compute_mode,
                    qt,
                    {col: fd.get_feature_store_format_version for col in (anchor_time(), window_end_column_name())},
                ).as_ref()
                renames = {
                    anchor_time(): DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME,
                    window_end_column_name(): DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME,
                    **partial_aggregate_column_renames(
                        slide_interval_string=fd.get_aggregate_slide_interval_string,
                        trailing_time_window_aggregation=fd.trailing_time_window_aggregation(),
                    ),
                }
                drop = []
            return RenameColsNode(dialect, compute_mode, qt, mapping=renames, drop=drop).as_ref()
        elif aggregation_level == AGGREGATION_LEVEL_DISABLED:
            return build_pipeline_querytree(
                dialect, compute_mode, fd, for_stream=False, feature_data_time_limits=feature_time_limits_aligned
            )
        elif aggregation_level == AGGREGATION_LEVEL_FULL:
            qt = build_aggregated_time_range_run_query(
                dialect,
                compute_mode,
                fd,
                feature_data_time_limits=feature_time_limits_aligned,
                aggregation_anchor_time=feature_end_time,
            )
            return qt

    msg = "Unsupported batch query tree"
    raise Exception(msg)


def _querytree_run_batch(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fd: FeatureDefinitionWrapper,
    feature_start_time: datetime,
    feature_end_time: datetime,
    mock_data_sources: Dict[str, NodeRef],
    aggregation_level: Optional[str],
) -> "tecton.framework.data_frame.DataFrame":
    feature_time_limits_aligned = pendulum.period(feature_start_time, feature_end_time)

    qt = _build_run_batch_querytree(
        dialect, compute_mode, fd, feature_end_time, feature_time_limits_aligned, aggregation_level
    )

    MockDataRewrite(mock_data_sources).rewrite(qt)

    return TectonDataFrame._create(qt)


def run_stream(
    fd: FeatureDefinitionWrapper,
    output_temp_table: str,
    checkpoint_dir: Optional[str],
) -> StreamingQuery:
    check_spark_version(fd.fv_spec.stream_cluster_config)
    plan = materialization_plan.get_stream_materialization_plan(
        spark=TectonContext.get_instance()._spark,
        feature_definition=fd,
    )
    spark_df = plan.online_store_data_frame
    with tempfile.TemporaryDirectory(dir=checkpoint_dir) as d:
        return (
            spark_df.writeStream.format("memory")
            .queryName(output_temp_table)
            .option("checkpointLocation", d)
            .outputMode("update")
            .start()
        )


def validate_run_transformation_args_match_transformation_mode(
    input_data: Dict[str, Union[Dict[str, Any], pandas.DataFrame, DataFrame, RealtimeContext]],
    transformation_mode: TransformationMode,
):
    for key, value in input_data.items():
        if transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
            if not isinstance(value, pandas.DataFrame) and not isinstance(value, DataFrame):
                msg = f"Expected input of type DataFrame for key '{key}' in 'pandas' mode, got {type(value).__name__}."
                raise errors.TectonValidationError(msg)
        elif transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYTHON:
            if not isinstance(value, Dict) and not isinstance(value, RealtimeContext):
                msg = f"Expected input of type Dict for key '{key}' in 'python' mode, got {type(value).__name__}."
                raise errors.TectonValidationError(msg)


def run_realtime(
    fd: FeatureDefinitionWrapper,
    fv_name: str,
    mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame, DataFrame, RealtimeContext]],
    transformation_mode: TransformationMode,
) -> Union[Dict[str, Any], "tecton.framework.data_frame.TectonDataFrame"]:
    validate_run_transformation_args_match_transformation_mode(mock_inputs, transformation_mode)
    for key in mock_inputs:
        if isinstance(mock_inputs[key], DataFrame):
            mock_inputs[key] = mock_inputs[key].toPandas()

    # Validate that all the mock_inputs match with FV inputs, and that num rows match across all mock_inputs.
    validate_realtime_mock_inputs_match_expected_shape(mock_inputs, fd.pipeline, fd)

    # Execute ODFV pipeline to get output DataFrame.
    output = run_mock_rtfv_pipeline(
        pipeline=fd.pipeline,
        transformations=fd.transformations,
        name=fv_name,
        mock_inputs=mock_inputs,
    )
    if isinstance(output, pandas.DataFrame):
        output = TectonDataFrame._create(output)

    return output


def run_mock_rtfv_pipeline(
    pipeline: Pipeline,
    transformations: List[specs.TransformationSpec],
    name: str,
    mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame]],
) -> Union[Dict[str, Any], pandas.DataFrame]:
    return RealtimeFeaturePipeline(
        name=name,
        pipeline=pipeline,
        transformations=transformations,
        udf_arg_idx_map={},
        output_schema=None,
        data_source_inputs=mock_inputs,
    ).to_dataframe()


def validate_and_get_aggregation_level(fd: FeatureDefinitionWrapper, aggregation_level: Optional[str]) -> str:
    # Set default aggregation_level value.
    if aggregation_level is None:
        if fd.is_temporal_aggregate:
            aggregation_level = AGGREGATION_LEVEL_FULL
        else:
            aggregation_level = AGGREGATION_LEVEL_DISABLED

    if aggregation_level not in SUPPORTED_AGGREGATION_LEVEL_VALUES:
        msg = "aggregation_level"
        raise errors.FV_INVALID_ARG_VALUE(msg, str(aggregation_level), str(SUPPORTED_AGGREGATION_LEVEL_VALUES))

    return aggregation_level


def _print_run_deprecation_message(aggregation_level: Optional[str]):
    if aggregation_level == AGGREGATION_LEVEL_FULL:
        logger.warning(errors.RUN_DEPRECATED_FULL_AGG)
    elif aggregation_level == AGGREGATION_LEVEL_DISABLED:
        logger.warning(errors.RUN_DEPRECATED_TRANSFORMATION)
    elif aggregation_level == AGGREGATION_LEVEL_PARTIAL:
        logger.warning(errors.RUN_DEPRECATED_PARTIAL_AGGS)


# Validate that mock_inputs keys are exact match with expected inputs.
def validate_realtime_mock_inputs_match_expected_shape(
    mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame, RealtimeContext]],
    pipeline: pipeline_pb2.Pipeline,
    rtfv_fd: Optional[FeatureDefinitionWrapper] = None,
):
    """Validate the mock_inputs for `run_transformation` matches expected input

    If a feature definition is passed, this will additionally check inputs against batch feature view schemas.
    """
    expected_input_names = get_all_input_keys(pipeline.root)
    mock_inputs_keys = set(mock_inputs.keys())
    if mock_inputs_keys != expected_input_names:
        raise errors.FV_INVALID_MOCK_INPUTS(list(mock_inputs_keys), list(expected_input_names))

    _validate_request_context_mock_inputs(mock_inputs, pipeline)
    # Get num row for all FV mock_inputs with DF types, to validate that they match.
    input_df_row_counts = set()
    for input in mock_inputs.values():
        if isinstance(input, pandas.DataFrame):
            input_df_row_counts.add(len(input.index))
    if len(input_df_row_counts) > 1:
        raise errors.FV_INVALID_MOCK_INPUTS_NUM_ROWS(input_df_row_counts)

    # Can only validate batch feature view inputs if the schema is known
    if rtfv_fd is not None:
        _validate_mock_bfv_inputs_to_odfv(mock_inputs, rtfv_fd)


def _validate_request_context_mock_inputs(
    mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame]],
    pipeline: pipeline_pb2.Pipeline,
):
    rc_node = pipeline_common.get_request_context_node(pipeline)
    if not rc_node:
        return

    expected_rc_schema = Schema(rc_node.request_context.tecton_schema)
    expected_cols = set(expected_rc_schema.column_names())
    actual_rc_inputs = mock_inputs[rc_node.input_name]

    # Python mode
    if isinstance(actual_rc_inputs, dict):
        actual_cols = set(actual_rc_inputs.keys())
    # Pandas mode
    elif isinstance(actual_rc_inputs, pandas.DataFrame):
        actual_cols = set(actual_rc_inputs.columns)

    if not actual_cols.issubset(expected_cols):
        raise errors.UNDEFINED_REQUEST_SOURCE_INPUT(list(actual_cols - expected_cols), list(expected_cols))


def _validate_mock_bfv_inputs_to_odfv(
    mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame]], odfv_fd: FeatureDefinitionWrapper
):
    """Validate the mock data used to represent batch feature view materialized data for `run()` and `test_run()`.

    Unlike other mock data checks, because there's no way for us to easily identify what features are needed in the
    ODFV, we expect all features in the upstream BFV to be present.
    """
    from tecton_core import feature_set_config

    # Extract input names (from pipeline) and map to feature view schemas (from FDW)
    dependent_fvs = feature_set_config.find_dependent_feature_set_items(
        odfv_fd.fco_container, odfv_fd.pipeline.root, visited_inputs={}, fv_id=odfv_fd.id
    )
    fv_ids_to_fvs = {fv.feature_definition.id: fv for fv in dependent_fvs}
    fco_ids_to_input_keys = get_fco_ids_to_input_keys(odfv_fd.pipeline.root)
    input_name_to_features = {}
    for fv_id, fv in fv_ids_to_fvs.items():
        input_name = fco_ids_to_input_keys[fv_id]
        input_name_to_features[input_name] = fv.features

    # Validate some required fields in the mock data schemas.
    for key, mock_df in mock_inputs.items():
        if key not in input_name_to_features:
            continue
        input_columns = mock_source_utils.get_pandas_or_spark_df_or_dict_columns(mock_df)
        # Check columns match BFV
        dependent_columns = input_name_to_features[key]
        unexpected_columns = [col for col in input_columns if col not in dependent_columns]
        if len(unexpected_columns) > 0:
            msg = f"Unexpected columns: {unexpected_columns} found in mock inputs. Expected columns from the Feature View {key}, such as: '{dependent_columns}'"
            raise TectonValidationError(msg)
        missing_columns = [col for col in dependent_columns if col not in input_columns]
        if len(missing_columns) > 0:
            logger.warning(
                f"ODFV {odfv_fd.name} has a dependency on the Feature View {key}. Features '{missing_columns}' of this "
                f"Feature View are not found. Available columns: '{input_columns}'"
            )



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Worker that receives input from Piped RDD.
"""
import os
import sys
import time
from inspect import getfullargspec
import importlib
# 'resource' is a Unix specific module.
has_resource_module = True
try:
    import resource
except ImportError:
    has_resource_module = False
import traceback

from pyspark.accumulators import _accumulatorRegistry
from pyspark.broadcast import Broadcast, _broadcastRegistry
from pyspark.java_gateway import local_connect_and_auth
from pyspark.taskcontext import BarrierTaskContext, TaskContext
from pyspark.files import SparkFiles
from pyspark.resource import ResourceInformation
from pyspark.rdd import PythonEvalType
from pyspark.serializers import write_with_length, write_int, read_long, read_bool, \
    write_long, read_int, SpecialLengths, UTF8Deserializer, PickleSerializer, \
    BatchedSerializer
from pyspark.sql.pandas.serializers import ArrowStreamPandasUDFSerializer, CogroupUDFSerializer
from pyspark.sql.pandas.types import to_arrow_type
from pyspark.sql.types import StructType
from pyspark.util import fail_on_stopiteration, try_simplify_traceback
from pyspark import shuffle

pickleSer = PickleSerializer()
utf8_deserializer = UTF8Deserializer()


def report_times(outfile, boot, init, finish):
    write_int(SpecialLengths.TIMING_DATA, outfile)
    write_long(int(1000 * boot), outfile)
    write_long(int(1000 * init), outfile)
    write_long(int(1000 * finish), outfile)


def add_path(path):
    # worker can be used, so do not add path multiple times
    if path not in sys.path:
        # overwrite system packages
        sys.path.insert(1, path)


def read_command(serializer, file):
    command = serializer._read_with_length(file)
    if isinstance(command, Broadcast):
        command = serializer.loads(command.value)
    return command


def chain(f, g):
    """chain two functions together """
    return lambda *a: g(f(*a))


def wrap_udf(f, return_type):
    if return_type.needConversion():
        toInternal = return_type.toInternal
        return lambda *a: toInternal(f(*a))
    else:
        return lambda *a: f(*a)


def wrap_scalar_pandas_udf(f, return_type):
    arrow_return_type = to_arrow_type(return_type)

    def verify_result_type(result):
        if not hasattr(result, "__len__"):
            pd_type = "Pandas.DataFrame" if type(return_type) == StructType else "Pandas.Series"
            raise TypeError("Return type of the user-defined function should be "
                            "{}, but is {}".format(pd_type, type(result)))
        return result

    def verify_result_length(result, length):
        if len(result) != length:
            raise RuntimeError("Result vector from pandas_udf was not the required length: "
                               "expected %d, got %d" % (length, len(result)))
        return result

    return lambda *a: (verify_result_length(
        verify_result_type(f(*a)), len(a[0])), arrow_return_type)


def wrap_pandas_iter_udf(f, return_type):
    arrow_return_type = to_arrow_type(return_type)

    def verify_result_type(result):
        if not hasattr(result, "__len__"):
            pd_type = "Pandas.DataFrame" if type(return_type) == StructType else "Pandas.Series"
            raise TypeError("Return type of the user-defined function should be "
                            "{}, but is {}".format(pd_type, type(result)))
        return result

    return lambda *iterator: map(lambda res: (res, arrow_return_type),
                                 map(verify_result_type, f(*iterator)))


def wrap_cogrouped_map_pandas_udf(f, return_type, argspec):

    def wrapped(left_key_series, left_value_series, right_key_series, right_value_series):
        import pandas as pd

        left_df = pd.concat(left_value_series, axis=1)
        right_df = pd.concat(right_value_series, axis=1)

        if len(argspec.args) == 2:
            result = f(left_df, right_df)
        elif len(argspec.args) == 3:
            key_series = left_key_series if not left_df.empty else right_key_series
            key = tuple(s[0] for s in key_series)
            result = f(key, left_df, right_df)
        if not isinstance(result, pd.DataFrame):
            raise TypeError("Return type of the user-defined function should be "
                            "pandas.DataFrame, but is {}".format(type(result)))
        if not len(result.columns) == len(return_type):
            raise RuntimeError(
                "Number of columns of the returned pandas.DataFrame "
                "doesn't match specified schema. "
                "Expected: {} Actual: {}".format(len(return_type), len(result.columns)))
        return result

    return lambda kl, vl, kr, vr: [(wrapped(kl, vl, kr, vr), to_arrow_type(return_type))]


def wrap_grouped_map_pandas_udf(f, return_type, argspec):

    def wrapped(key_series, value_series):
        import pandas as pd

        if len(argspec.args) == 1:
            result = f(pd.concat(value_series, axis=1))
        elif len(argspec.args) == 2:
            key = tuple(s[0] for s in key_series)
            result = f(key, pd.concat(value_series, axis=1))

        if not isinstance(result, pd.DataFrame):
            raise TypeError("Return type of the user-defined function should be "
                            "pandas.DataFrame, but is {}".format(type(result)))
        if not len(result.columns) == len(return_type):
            raise RuntimeError(
                "Number of columns of the returned pandas.DataFrame "
                "doesn't match specified schema. "
                "Expected: {} Actual: {}".format(len(return_type), len(result.columns)))
        return result

    return lambda k, v: [(wrapped(k, v), to_arrow_type(return_type))]


def wrap_grouped_agg_pandas_udf(f, return_type):
    arrow_return_type = to_arrow_type(return_type)

    def wrapped(*series):
        import pandas as pd
        result = f(*series)
        return pd.Series([result])

    return lambda *a: (wrapped(*a), arrow_return_type)


def wrap_window_agg_pandas_udf(f, return_type, runner_conf, udf_index):
    window_bound_types_str = runner_conf.get('pandas_window_bound_types')
    window_bound_type = [t.strip().lower() for t in window_bound_types_str.split(',')][udf_index]
    if window_bound_type == 'bounded':
        return wrap_bounded_window_agg_pandas_udf(f, return_type)
    elif window_bound_type == 'unbounded':
        return wrap_unbounded_window_agg_pandas_udf(f, return_type)
    else:
        raise RuntimeError("Invalid window bound type: {} ".format(window_bound_type))


def wrap_unbounded_window_agg_pandas_udf(f, return_type):
    # This is similar to grouped_agg_pandas_udf, the only difference
    # is that window_agg_pandas_udf needs to repeat the return value
    # to match window length, where grouped_agg_pandas_udf just returns
    # the scalar value.
    arrow_return_type = to_arrow_type(return_type)

    def wrapped(*series):
        import pandas as pd
        result = f(*series)
        return pd.Series([result]).repeat(len(series[0]))

    return lambda *a: (wrapped(*a), arrow_return_type)


def wrap_bounded_window_agg_pandas_udf(f, return_type):
    arrow_return_type = to_arrow_type(return_type)

    def wrapped(begin_index, end_index, *series):
        import pandas as pd
        result = []

        # Index operation is faster on np.ndarray,
        # So we turn the index series into np array
        # here for performance
        begin_array = begin_index.values
        end_array = end_index.values

        for i in range(len(begin_array)):
            # Note: Create a slice from a series for each window is
            #       actually pretty expensive. However, there
            #       is no easy way to reduce cost here.
            # Note: s.iloc[i : j] is about 30% faster than s[i: j], with
            #       the caveat that the created slices shares the same
            #       memory with s. Therefore, user are not allowed to
            #       change the value of input series inside the window
            #       function. It is rare that user needs to modify the
            #       input series in the window function, and therefore,
            #       it is be a reasonable restriction.
            # Note: Calling reset_index on the slices will increase the cost
            #       of creating slices by about 100%. Therefore, for performance
            #       reasons we don't do it here.
            series_slices = [s.iloc[begin_array[i]: end_array[i]] for s in series]
            result.append(f(*series_slices))
        return pd.Series(result)

    return lambda *a: (wrapped(*a), arrow_return_type)


def read_single_udf(pickleSer, infile, eval_type, runner_conf, udf_index):
    num_arg = read_int(infile)
    arg_offsets = [read_int(infile) for i in range(num_arg)]
    chained_func = None
    for i in range(read_int(infile)):
        f, return_type = read_command(pickleSer, infile)
        if chained_func is None:
            chained_func = f
        else:
            chained_func = chain(chained_func, f)

    if eval_type == PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF:
        func = chained_func
    else:
        # make sure StopIteration's raised in the user code are not ignored
        # when they are processed in a for loop, raise them as RuntimeError's instead
        func = fail_on_stopiteration(chained_func)

    # the last returnType will be the return type of UDF
    if eval_type == PythonEvalType.SQL_SCALAR_PANDAS_UDF:
        return arg_offsets, wrap_scalar_pandas_udf(func, return_type)
    elif eval_type == PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF:
        return arg_offsets, wrap_pandas_iter_udf(func, return_type)
    elif eval_type == PythonEvalType.SQL_MAP_PANDAS_ITER_UDF:
        return arg_offsets, wrap_pandas_iter_udf(func, return_type)
    elif eval_type == PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF:
        argspec = getfullargspec(chained_func)  # signature was lost when wrapping it
        return arg_offsets, wrap_grouped_map_pandas_udf(func, return_type, argspec)
    elif eval_type == PythonEvalType.SQL_COGROUPED_MAP_PANDAS_UDF:
        argspec = getfullargspec(chained_func)  # signature was lost when wrapping it
        return arg_offsets, wrap_cogrouped_map_pandas_udf(func, return_type, argspec)
    elif eval_type == PythonEvalType.SQL_GROUPED_AGG_PANDAS_UDF:
        return arg_offsets, wrap_grouped_agg_pandas_udf(func, return_type)
    elif eval_type == PythonEvalType.SQL_WINDOW_AGG_PANDAS_UDF:
        return arg_offsets, wrap_window_agg_pandas_udf(func, return_type, runner_conf, udf_index)
    elif eval_type == PythonEvalType.SQL_BATCHED_UDF:
        return arg_offsets, wrap_udf(func, return_type)
    else:
        raise ValueError("Unknown eval type: {}".format(eval_type))


def read_udfs(pickleSer, infile, eval_type):
    runner_conf = {}

    if eval_type in (PythonEvalType.SQL_SCALAR_PANDAS_UDF,
                     PythonEvalType.SQL_COGROUPED_MAP_PANDAS_UDF,
                     PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF,
                     PythonEvalType.SQL_MAP_PANDAS_ITER_UDF,
                     PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF,
                     PythonEvalType.SQL_GROUPED_AGG_PANDAS_UDF,
                     PythonEvalType.SQL_WINDOW_AGG_PANDAS_UDF):

        # Load conf used for pandas_udf evaluation
        num_conf = read_int(infile)
        for i in range(num_conf):
            k = utf8_deserializer.loads(infile)
            v = utf8_deserializer.loads(infile)
            runner_conf[k] = v

        # NOTE: if timezone is set here, that implies respectSessionTimeZone is True
        timezone = runner_conf.get("spark.sql.session.timeZone", None)
        safecheck = runner_conf.get("spark.sql.execution.pandas.convertToArrowArraySafely",
                                    "false").lower() == 'true'
        # Used by SQL_GROUPED_MAP_PANDAS_UDF and SQL_SCALAR_PANDAS_UDF when returning StructType
        assign_cols_by_name = runner_conf.get(
            "spark.sql.legacy.execution.pandas.groupedMap.assignColumnsByName", "true")\
            .lower() == "true"

        if eval_type == PythonEvalType.SQL_COGROUPED_MAP_PANDAS_UDF:
            ser = CogroupUDFSerializer(timezone, safecheck, assign_cols_by_name)
        else:
            # Scalar Pandas UDF handles struct type arguments as pandas DataFrames instead of
            # pandas Series. See SPARK-27240.
            df_for_struct = (eval_type == PythonEvalType.SQL_SCALAR_PANDAS_UDF or
                             eval_type == PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF or
                             eval_type == PythonEvalType.SQL_MAP_PANDAS_ITER_UDF)
            ser = ArrowStreamPandasUDFSerializer(timezone, safecheck, assign_cols_by_name,
                                                 df_for_struct)
    else:
        ser = BatchedSerializer(PickleSerializer(), 100)

    num_udfs = read_int(infile)

    is_scalar_iter = eval_type == PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF
    is_map_iter = eval_type == PythonEvalType.SQL_MAP_PANDAS_ITER_UDF

    if is_scalar_iter or is_map_iter:
        if is_scalar_iter:
            assert num_udfs == 1, "One SCALAR_ITER UDF expected here."
        if is_map_iter:
            assert num_udfs == 1, "One MAP_ITER UDF expected here."

        arg_offsets, udf = read_single_udf(
            pickleSer, infile, eval_type, runner_conf, udf_index=0)

        def func(_, iterator):
            num_input_rows = 0

            def map_batch(batch):
                nonlocal num_input_rows

                udf_args = [batch[offset] for offset in arg_offsets]
                num_input_rows += len(udf_args[0])
                if len(udf_args) == 1:
                    return udf_args[0]
                else:
                    return tuple(udf_args)

            iterator = map(map_batch, iterator)
            result_iter = udf(iterator)

            num_output_rows = 0
            for result_batch, result_type in result_iter:
                num_output_rows += len(result_batch)
                # This assert is for Scalar Iterator UDF to fail fast.
                # The length of the entire input can only be explicitly known
                # by consuming the input iterator in user side. Therefore,
                # it's very unlikely the output length is higher than
                # input length.
                assert is_map_iter or num_output_rows <= num_input_rows, \
                    "Pandas SCALAR_ITER UDF outputted more rows than input rows."
                yield (result_batch, result_type)

            if is_scalar_iter:
                try:
                    next(iterator)
                except StopIteration:
                    pass
                else:
                    raise RuntimeError("pandas iterator UDF should exhaust the input "
                                       "iterator.")

                if num_output_rows != num_input_rows:
                    raise RuntimeError(
                        "The length of output in Scalar iterator pandas UDF should be "
                        "the same with the input's; however, the length of output was %d and the "
                        "length of input was %d." % (num_output_rows, num_input_rows))

        # profiling is not supported for UDF
        return func, None, ser, ser

    def extract_key_value_indexes(grouped_arg_offsets):
        """
        Helper function to extract the key and value indexes from arg_offsets for the grouped and
        cogrouped pandas udfs. See BasePandasGroupExec.resolveArgOffsets for equivalent scala code.

        Parameters
        ----------
        grouped_arg_offsets:  list
            List containing the key and value indexes of columns of the
            DataFrames to be passed to the udf. It consists of n repeating groups where n is the
            number of DataFrames.  Each group has the following format:
                group[0]: length of group
                group[1]: length of key indexes
                group[2.. group[1] +2]: key attributes
                group[group[1] +3 group[0]]: value attributes
        """
        parsed = []
        idx = 0
        while idx < len(grouped_arg_offsets):
            offsets_len = grouped_arg_offsets[idx]
            idx += 1
            offsets = grouped_arg_offsets[idx: idx + offsets_len]
            split_index = offsets[0] + 1
            offset_keys = offsets[1: split_index]
            offset_values = offsets[split_index:]
            parsed.append([offset_keys, offset_values])
            idx += offsets_len
        return parsed

    if eval_type == PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF:
        # We assume there is only one UDF here because grouped map doesn't
        # support combining multiple UDFs.
        assert num_udfs == 1

        # See FlatMapGroupsInPandasExec for how arg_offsets are used to
        # distinguish between grouping attributes and data attributes
        arg_offsets, f = read_single_udf(pickleSer, infile, eval_type, runner_conf, udf_index=0)
        parsed_offsets = extract_key_value_indexes(arg_offsets)

        # Create function like this:
        #   mapper a: f([a[0]], [a[0], a[1]])
        def mapper(a):
            keys = [a[o] for o in parsed_offsets[0][0]]
            vals = [a[o] for o in parsed_offsets[0][1]]
            return f(keys, vals)
    elif eval_type == PythonEvalType.SQL_COGROUPED_MAP_PANDAS_UDF:
        # We assume there is only one UDF here because cogrouped map doesn't
        # support combining multiple UDFs.
        assert num_udfs == 1
        arg_offsets, f = read_single_udf(pickleSer, infile, eval_type, runner_conf, udf_index=0)

        parsed_offsets = extract_key_value_indexes(arg_offsets)

        def mapper(a):
            df1_keys = [a[0][o] for o in parsed_offsets[0][0]]
            df1_vals = [a[0][o] for o in parsed_offsets[0][1]]
            df2_keys = [a[1][o] for o in parsed_offsets[1][0]]
            df2_vals = [a[1][o] for o in parsed_offsets[1][1]]
            return f(df1_keys, df1_vals, df2_keys, df2_vals)
    else:
        udfs = []
        for i in range(num_udfs):
            udfs.append(read_single_udf(pickleSer, infile, eval_type, runner_conf, udf_index=i))

        def mapper(a):
            result = tuple(f(*[a[o] for o in arg_offsets]) for (arg_offsets, f) in udfs)
            # In the special case of a single UDF this will return a single result rather
            # than a tuple of results; this is the format that the JVM side expects.
            if len(result) == 1:
                return result[0]
            else:
                return result

    func = lambda _, it: map(mapper, it)

    # profiling is not supported for UDF
    return func, None, ser, ser


def main(infile, outfile):
    try:
        boot_time = time.time()
        split_index = read_int(infile)
        if split_index == -1:  # for unit tests
            sys.exit(-1)

        version = utf8_deserializer.loads(infile)
        if version != "%d.%d" % sys.version_info[:2]:
            raise Exception(("Python in worker has different version %s than that in " +
                             "driver %s, PySpark cannot run with different minor versions. " +
                             "Please check environment variables PYSPARK_PYTHON and " +
                             "PYSPARK_DRIVER_PYTHON are correctly set.") %
                            ("%d.%d" % sys.version_info[:2], version))

        # read inputs only for a barrier task
        isBarrier = read_bool(infile)
        boundPort = read_int(infile)
        secret = UTF8Deserializer().loads(infile)

        # set up memory limits
        memory_limit_mb = int(os.environ.get('PYSPARK_EXECUTOR_MEMORY_MB', "-1"))
        if memory_limit_mb > 0 and has_resource_module:
            total_memory = resource.RLIMIT_AS
            try:
                (soft_limit, hard_limit) = resource.getrlimit(total_memory)
                msg = "Current mem limits: {0} of max {1}\n".format(soft_limit, hard_limit)
                print(msg, file=sys.stderr)

                # convert to bytes
                new_limit = memory_limit_mb * 1024 * 1024

                if soft_limit == resource.RLIM_INFINITY or new_limit < soft_limit:
                    msg = "Setting mem limits to {0} of max {1}\n".format(new_limit, new_limit)
                    print(msg, file=sys.stderr)
                    resource.setrlimit(total_memory, (new_limit, new_limit))

            except (resource.error, OSError, ValueError) as e:
                # not all systems support resource limits, so warn instead of failing
                print("WARN: Failed to set memory limit: {0}\n".format(e), file=sys.stderr)

        # initialize global state
        taskContext = None
        if isBarrier:
            taskContext = BarrierTaskContext._getOrCreate()
            BarrierTaskContext._initialize(boundPort, secret)
            # Set the task context instance here, so we can get it by TaskContext.get for
            # both TaskContext and BarrierTaskContext
            TaskContext._setTaskContext(taskContext)
        else:
            taskContext = TaskContext._getOrCreate()
        # read inputs for TaskContext info
        taskContext._stageId = read_int(infile)
        taskContext._partitionId = read_int(infile)
        taskContext._attemptNumber = read_int(infile)
        taskContext._taskAttemptId = read_long(infile)
        taskContext._resources = {}
        for r in range(read_int(infile)):
            key = utf8_deserializer.loads(infile)
            name = utf8_deserializer.loads(infile)
            addresses = []
            taskContext._resources = {}
            for a in range(read_int(infile)):
                addresses.append(utf8_deserializer.loads(infile))
            taskContext._resources[key] = ResourceInformation(name, addresses)

        taskContext._localProperties = dict()
        for i in range(read_int(infile)):
            k = utf8_deserializer.loads(infile)
            v = utf8_deserializer.loads(infile)
            taskContext._localProperties[k] = v

        shuffle.MemoryBytesSpilled = 0
        shuffle.DiskBytesSpilled = 0
        _accumulatorRegistry.clear()

        # fetch name of workdir
        spark_files_dir = utf8_deserializer.loads(infile)
        SparkFiles._root_directory = spark_files_dir
        SparkFiles._is_running_on_worker = True

        # fetch names of includes (*.zip and *.egg files) and construct PYTHONPATH
        add_path(spark_files_dir)  # *.py files that were added will be copied here
        num_python_includes = read_int(infile)
        for _ in range(num_python_includes):
            filename = utf8_deserializer.loads(infile)
            add_path(os.path.join(spark_files_dir, filename))

        importlib.invalidate_caches()

        # fetch names and values of broadcast variables
        needs_broadcast_decryption_server = read_bool(infile)
        num_broadcast_variables = read_int(infile)
        if needs_broadcast_decryption_server:
            # read the decrypted data from a server in the jvm
            port = read_int(infile)
            auth_secret = utf8_deserializer.loads(infile)
            (broadcast_sock_file, _) = local_connect_and_auth(port, auth_secret)

        for _ in range(num_broadcast_variables):
            bid = read_long(infile)
            if bid >= 0:
                if needs_broadcast_decryption_server:
                    read_bid = read_long(broadcast_sock_file)
                    assert(read_bid == bid)
                    _broadcastRegistry[bid] = \
                        Broadcast(sock_file=broadcast_sock_file)
                else:
                    path = utf8_deserializer.loads(infile)
                    _broadcastRegistry[bid] = Broadcast(path=path)

            else:
                bid = - bid - 1
                _broadcastRegistry.pop(bid)

        if needs_broadcast_decryption_server:
            broadcast_sock_file.write(b'1')
            broadcast_sock_file.close()

        _accumulatorRegistry.clear()
        eval_type = read_int(infile)
        if eval_type == PythonEvalType.NON_UDF:
            func, profiler, deserializer, serializer = read_command(pickleSer, infile)
        else:
            func, profiler, deserializer, serializer = read_udfs(pickleSer, infile, eval_type)

        init_time = time.time()

        def process():
            iterator = deserializer.load_stream(infile)
            out_iter = func(split_index, iterator)
            try:
                serializer.dump_stream(out_iter, outfile)
            finally:
                if hasattr(out_iter, 'close'):
                    out_iter.close()

        if profiler:
            profiler.profile(process)
        else:
            process()

        # Reset task context to None. This is a guard code to avoid residual context when worker
        # reuse.
        TaskContext._setTaskContext(None)
        BarrierTaskContext._setTaskContext(None)
    except BaseException as e:
        try:
            exc_info = None
            if os.environ.get("SPARK_SIMPLIFIED_TRACEBACK", False):
                tb = try_simplify_traceback(sys.exc_info()[-1])
                if tb is not None:
                    e.__cause__ = None
                    exc_info = "".join(traceback.format_exception(type(e), e, tb))
            if exc_info is None:
                exc_info = traceback.format_exc()

            write_int(SpecialLengths.PYTHON_EXCEPTION_THROWN, outfile)
            write_with_length(exc_info.encode("utf-8"), outfile)
        except IOError:
            # JVM close the socket
            pass
        except BaseException:
            # Write the error to stderr if it happened while serializing
            print("PySpark worker failed with exception:", file=sys.stderr)
            print(traceback.format_exc(), file=sys.stderr)
        sys.exit(-1)
    finish_time = time.time()
    report_times(outfile, boot_time, init_time, finish_time)
    write_long(shuffle.MemoryBytesSpilled, outfile)
    write_long(shuffle.DiskBytesSpilled, outfile)

    # Mark the beginning of the accumulators section of the output
    write_int(SpecialLengths.END_OF_DATA_SECTION, outfile)
    write_int(len(_accumulatorRegistry), outfile)
    for (aid, accum) in _accumulatorRegistry.items():
        pickleSer._write_with_length((aid, accum._value), outfile)

    # check end of stream
    if read_int(infile) == SpecialLengths.END_OF_STREAM:
        write_int(SpecialLengths.END_OF_STREAM, outfile)
    else:
        # write a different value to tell JVM to not reuse this worker
        write_int(SpecialLengths.END_OF_DATA_SECTION, outfile)
        sys.exit(-1)


if __name__ == '__main__':
    # Read information about how to connect back to the JVM from the environment.
    java_port = int(os.environ["PYTHON_WORKER_FACTORY_PORT"])
    auth_secret = os.environ["PYTHON_WORKER_FACTORY_SECRET"]
    (sock_file, _) = local_connect_and_auth(java_port, auth_secret)
    main(sock_file, sock_file)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/materialization/job_metadata__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.spark_common import clusters__client_pb2 as tecton__proto_dot_spark__common_dot_clusters____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7tecton_proto/materialization/job_metadata__client.proto\x12\x1ctecton_proto.materialization\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x30tecton_proto/spark_common/clusters__client.proto\"\xfb\x03\n\x0bJobMetadata\x12\x8a\x01\n\"online_store_copier_execution_info\x18\x01 \x01(\x0b\x32<.tecton_proto.materialization.OnlineStoreCopierExecutionInfoH\x00R\x1eonlineStoreCopierExecutionInfo\x12g\n\x14spark_execution_info\x18\x03 \x01(\x0b\x32\x33.tecton_proto.materialization.SparkJobExecutionInfoH\x00R\x12sparkExecutionInfo\x12\x61\n\x13tecton_managed_info\x18\x04 \x01(\x0b\x32/.tecton_proto.materialization.TectonManagedInfoH\x00R\x11tectonManagedInfo\x12\x86\x01\n materialization_consumption_info\x18\x02 \x01(\x0b\x32<.tecton_proto.materialization.MaterializationConsumptionInfoR\x1ematerializationConsumptionInfoB\n\n\x08job_info\"?\n\x1eOnlineStoreCopierExecutionInfo\x12\x1d\n\nis_revoked\x18\x01 \x01(\x08R\tisRevoked\"\xdd\x01\n\x15SparkJobExecutionInfo\x12\x15\n\x06run_id\x18\x01 \x01(\tR\x05runId\x12\x1d\n\nis_revoked\x18\x02 \x01(\x08R\tisRevoked\x12\x8d\x01\n#stream_handoff_synchronization_info\x18\x03 \x01(\x0b\x32>.tecton_proto.materialization.StreamHandoffSynchronizationInfoR streamHandoffSynchronizationInfo\"\x91\x02\n StreamHandoffSynchronizationInfo\x12.\n\x13new_cluster_started\x18\x01 \x01(\x08R\x11newClusterStarted\x12;\n\x1astream_query_start_allowed\x18\x02 \x01(\x08R\x17streamQueryStartAllowed\x12@\n\x1cquery_cancellation_requested\x18\x03 \x01(\x08R\x1aqueryCancellationRequested\x12>\n\x1bquery_cancellation_complete\x18\x04 \x01(\x08R\x19queryCancellationComplete\"\xfc\x02\n\x1eMaterializationConsumptionInfo\x12z\n\x19offline_store_consumption\x18\x01 \x01(\x0b\x32>.tecton_proto.materialization.OfflineStoreWriteConsumptionInfoR\x17offlineStoreConsumption\x12w\n\x18online_store_consumption\x18\x02 \x01(\x0b\x32=.tecton_proto.materialization.OnlineStoreWriteConsumptionInfoR\x16onlineStoreConsumption\x12\x65\n\x13\x63ompute_consumption\x18\x03 \x01(\x0b\x32\x34.tecton_proto.materialization.ComputeConsumptionInfoR\x12\x63omputeConsumption\"\xfc\x02\n OfflineStoreWriteConsumptionInfo\x12~\n\x10\x63onsumption_info\x18\x01 \x03(\x0b\x32S.tecton_proto.materialization.OfflineStoreWriteConsumptionInfo.ConsumptionInfoEntryR\x0f\x63onsumptionInfo\x12\\\n\x12offline_store_type\x18\x02 \x01(\x0e\x32..tecton_proto.materialization.OfflineStoreTypeR\x10offlineStoreType\x1az\n\x14\x43onsumptionInfoEntry\x12\x10\n\x03key\x18\x01 \x01(\x03R\x03key\x12L\n\x05value\x18\x02 \x01(\x0b\x32\x36.tecton_proto.materialization.OfflineConsumptionBucketR\x05value:\x02\x38\x01\"h\n\x18OfflineConsumptionBucket\x12!\n\x0crows_written\x18\x01 \x01(\x03R\x0browsWritten\x12)\n\x10\x66\x65\x61tures_written\x18\x02 \x01(\x03R\x0f\x66\x65\x61turesWritten\"\xf6\x02\n\x1fOnlineStoreWriteConsumptionInfo\x12}\n\x10\x63onsumption_info\x18\x01 \x03(\x0b\x32R.tecton_proto.materialization.OnlineStoreWriteConsumptionInfo.ConsumptionInfoEntryR\x0f\x63onsumptionInfo\x12Y\n\x11online_store_type\x18\x02 \x01(\x0e\x32-.tecton_proto.materialization.OnlineStoreTypeR\x0fonlineStoreType\x1ay\n\x14\x43onsumptionInfoEntry\x12\x10\n\x03key\x18\x01 \x01(\x03R\x03key\x12K\n\x05value\x18\x02 \x01(\x0b\x32\x35.tecton_proto.materialization.OnlineConsumptionBucketR\x05value:\x02\x38\x01\"m\n\x17OnlineConsumptionBucket\x12!\n\x0crows_written\x18\x01 \x01(\x03R\x0browsWritten\x12)\n\x10\x66\x65\x61tures_written\x18\x02 \x01(\x03R\x0f\x66\x65\x61turesWrittenJ\x04\x08\x03\x10\x04\"\xa0\x01\n\x16\x43omputeConsumptionInfo\x12\x35\n\x08\x64uration\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x08\x64uration\x12O\n\rcompute_usage\x18\x02 \x03(\x0b\x32*.tecton_proto.materialization.ComputeUsageR\x0c\x63omputeUsage\"\xbb\x01\n\x0c\x43omputeUsage\x12_\n\x15instance_availability\x18\x01 \x01(\x0e\x32*.tecton_proto.spark_common.AwsAvailabilityR\x14instanceAvailability\x12#\n\rinstance_type\x18\x02 \x01(\tR\x0cinstanceType\x12%\n\x0einstance_count\x18\x03 \x01(\x03R\rinstanceCount\"\xd8\x06\n\x12TectonManagedStage\x12Y\n\nstage_type\x18\x01 \x01(\x0e\x32:.tecton_proto.materialization.TectonManagedStage.StageTypeR\tstageType\x12L\n\x05state\x18\x02 \x01(\x0e\x32\x36.tecton_proto.materialization.TectonManagedStage.StateR\x05state\x12#\n\rexternal_link\x18\x03 \x01(\tR\x0c\x65xternalLink\x12\x1a\n\x08progress\x18\x04 \x01(\x02R\x08progress\x12 \n\x0b\x64\x65scription\x18\x05 \x01(\tR\x0b\x64\x65scription\x12Y\n\nerror_type\x18\x06 \x01(\x0e\x32:.tecton_proto.materialization.TectonManagedStage.ErrorTypeR\terrorType\x12!\n\x0c\x65rror_detail\x18\x07 \x01(\tR\x0b\x65rrorDetail\x12,\n\x12\x63ompiled_sql_query\x18\x08 \x01(\tR\x10\x63ompiledSqlQuery\x12\x35\n\x08\x64uration\x18\t \x01(\x0b\x32\x19.google.protobuf.DurationR\x08\x64uration\x12\x39\n\nstart_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tstartTime\"h\n\tStageType\x12\r\n\tSNOWFLAKE\x10\x01\x12\n\n\x06PYTHON\x10\x02\x12\r\n\tAGGREGATE\x10\x03\x12\x11\n\rOFFLINE_STORE\x10\x04\x12\x10\n\x0cONLINE_STORE\x10\x05\x12\x0c\n\x08\x42IGQUERY\x10\x06\"_\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07PENDING\x10\x01\x12\x0b\n\x07RUNNING\x10\x02\x12\x0b\n\x07SUCCESS\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\r\n\tCANCELLED\x10\x05\"M\n\tErrorType\x12\x1a\n\x16\x45RROR_TYPE_UNSPECIFIED\x10\x00\x12\x14\n\x10UNEXPECTED_ERROR\x10\x01\x12\x0e\n\nUSER_ERROR\x10\x03\"\xab\x01\n\x11TectonManagedInfo\x12H\n\x06stages\x18\x01 \x03(\x0b\x32\x30.tecton_proto.materialization.TectonManagedStageR\x06stages\x12L\n\x05state\x18\x02 \x01(\x0e\x32\x36.tecton_proto.materialization.TectonManagedStage.StateR\x05state*\x8f\x01\n\x0fOnlineStoreType\x12!\n\x1dONLINE_STORE_TYPE_UNSPECIFIED\x10\x00\x12\x1c\n\x18ONLINE_STORE_TYPE_DYNAMO\x10\x01\x12\x1b\n\x17ONLINE_STORE_TYPE_REDIS\x10\x02\x12\x1e\n\x1aONLINE_STORE_TYPE_BIGTABLE\x10\x03*\x8a\x01\n\x10OfflineStoreType\x12\"\n\x1eOFFLINE_STORE_TYPE_UNSPECIFIED\x10\x00\x12\x19\n\x15OFFLINE_STORE_TYPE_S3\x10\x01\x12\x1b\n\x17OFFLINE_STORE_TYPE_DBFS\x10\x02\x12\x1a\n\x16OFFLINE_STORE_TYPE_GCS\x10\x03*\x84\x01\n\x14JobMetadataTableType\x12\'\n#JOB_METADATA_TABLE_TYPE_UNSPECIFIED\x10\x00\x12\"\n\x1eJOB_METADATA_TABLE_TYPE_DYNAMO\x10\x01\x12\x1f\n\x1bJOB_METADATA_TABLE_TYPE_GCS\x10\x02\x42\x1e\n\x1a\x63om.tecton.materializationP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.materialization.job_metadata__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.tecton.materializationP\001'
  _OFFLINESTOREWRITECONSUMPTIONINFO_CONSUMPTIONINFOENTRY._options = None
  _OFFLINESTOREWRITECONSUMPTIONINFO_CONSUMPTIONINFOENTRY._serialized_options = b'8\001'
  _ONLINESTOREWRITECONSUMPTIONINFO_CONSUMPTIONINFOENTRY._options = None
  _ONLINESTOREWRITECONSUMPTIONINFO_CONSUMPTIONINFOENTRY._serialized_options = b'8\001'
  _ONLINESTORETYPE._serialized_start=4026
  _ONLINESTORETYPE._serialized_end=4169
  _OFFLINESTORETYPE._serialized_start=4172
  _OFFLINESTORETYPE._serialized_end=4310
  _JOBMETADATATABLETYPE._serialized_start=4313
  _JOBMETADATATABLETYPE._serialized_end=4445
  _JOBMETADATA._serialized_start=205
  _JOBMETADATA._serialized_end=712
  _ONLINESTORECOPIEREXECUTIONINFO._serialized_start=714
  _ONLINESTORECOPIEREXECUTIONINFO._serialized_end=777
  _SPARKJOBEXECUTIONINFO._serialized_start=780
  _SPARKJOBEXECUTIONINFO._serialized_end=1001
  _STREAMHANDOFFSYNCHRONIZATIONINFO._serialized_start=1004
  _STREAMHANDOFFSYNCHRONIZATIONINFO._serialized_end=1277
  _MATERIALIZATIONCONSUMPTIONINFO._serialized_start=1280
  _MATERIALIZATIONCONSUMPTIONINFO._serialized_end=1660
  _OFFLINESTOREWRITECONSUMPTIONINFO._serialized_start=1663
  _OFFLINESTOREWRITECONSUMPTIONINFO._serialized_end=2043
  _OFFLINESTOREWRITECONSUMPTIONINFO_CONSUMPTIONINFOENTRY._serialized_start=1921
  _OFFLINESTOREWRITECONSUMPTIONINFO_CONSUMPTIONINFOENTRY._serialized_end=2043
  _OFFLINECONSUMPTIONBUCKET._serialized_start=2045
  _OFFLINECONSUMPTIONBUCKET._serialized_end=2149
  _ONLINESTOREWRITECONSUMPTIONINFO._serialized_start=2152
  _ONLINESTOREWRITECONSUMPTIONINFO._serialized_end=2526
  _ONLINESTOREWRITECONSUMPTIONINFO_CONSUMPTIONINFOENTRY._serialized_start=2405
  _ONLINESTOREWRITECONSUMPTIONINFO_CONSUMPTIONINFOENTRY._serialized_end=2526
  _ONLINECONSUMPTIONBUCKET._serialized_start=2528
  _ONLINECONSUMPTIONBUCKET._serialized_end=2637
  _COMPUTECONSUMPTIONINFO._serialized_start=2640
  _COMPUTECONSUMPTIONINFO._serialized_end=2800
  _COMPUTEUSAGE._serialized_start=2803
  _COMPUTEUSAGE._serialized_end=2990
  _TECTONMANAGEDSTAGE._serialized_start=2993
  _TECTONMANAGEDSTAGE._serialized_end=3849
  _TECTONMANAGEDSTAGE_STAGETYPE._serialized_start=3569
  _TECTONMANAGEDSTAGE_STAGETYPE._serialized_end=3673
  _TECTONMANAGEDSTAGE_STATE._serialized_start=3675
  _TECTONMANAGEDSTAGE_STATE._serialized_end=3770
  _TECTONMANAGEDSTAGE_ERRORTYPE._serialized_start=3772
  _TECTONMANAGEDSTAGE_ERRORTYPE._serialized_end=3849
  _TECTONMANAGEDINFO._serialized_start=3852
  _TECTONMANAGEDINFO._serialized_end=4023
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/state_update__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.args import diff_options__client_pb2 as tecton__proto_dot_args_dot_diff__options____client__pb2
from tecton_proto.args import fco_args__client_pb2 as tecton__proto_dot_args_dot_fco__args____client__pb2
from tecton_proto.args import repo_metadata__client_pb2 as tecton__proto_dot_args_dot_repo__metadata____client__pb2
from tecton_proto.auth import principal__client_pb2 as tecton__proto_dot_auth_dot_principal____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,tecton_proto/data/state_update__client.proto\x12\x11tecton_proto.data\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a,tecton_proto/args/diff_options__client.proto\x1a(tecton_proto/args/fco_args__client.proto\x1a-tecton_proto/args/repo_metadata__client.proto\x1a)tecton_proto/auth/principal__client.proto\x1a$tecton_proto/common/id__client.proto\"\xa2\x02\n\x0f\x46\x63oPropertyDiff\x12#\n\rproperty_name\x18\x01 \x01(\tR\x0cpropertyName\x12!\n\x0cval_existing\x18\x02 \x01(\tR\x0bvalExisting\x12!\n\x0cval_declared\x18\x03 \x01(\tR\x0bvalDeclared\x12R\n\x0erendering_type\x18\x04 \x01(\x0e\x32+.tecton_proto.args.FcoPropertyRenderingTypeR\rrenderingType\x12P\n\x11\x63ustom_comparator\x18\x05 \x01(\x0e\x32#.tecton_proto.args.CustomComparatorR\x10\x63ustomComparator\"\x87\x01\n\x18\x46\x63oTransitionSideEffects\x12k\n\x13stream_restart_type\x18\x01 \x01(\x0e\x32;.tecton_proto.data.FcoTransitionSideEffectStreamRestartTypeR\x11streamRestartType\"\xe6\x01\n\x1c\x42\x61tchMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\x12\x46\n\x11schedule_interval\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x10scheduleInterval\x12W\n\x0b\x64\x65stination\x18\x03 \x01(\x0e\x32\x35.tecton_proto.data.MaterializationTaskDiffDestinationR\x0b\x64\x65stination\"F\n\x1dStreamMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\"\xd7\x02\n\x1f\x42\x61\x63kfillMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\x12H\n\x12\x66\x65\x61ture_start_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x66\x65\x61tureStartTime\x12\x44\n\x10\x66\x65\x61ture_end_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x66\x65\x61tureEndTime\x12$\n\x0enumber_of_jobs\x18\x04 \x01(\x05R\x0cnumberOfJobs\x12W\n\x0b\x64\x65stination\x18\x05 \x01(\x0e\x32\x35.tecton_proto.data.MaterializationTaskDiffDestinationR\x0b\x64\x65stination\"D\n\x1bPlanIntegrationTestTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\"\x8f\x04\n\x13MaterializationInfo\x12\x62\n\x13\x62\x61\x63kfill_task_diffs\x18\x01 \x03(\x0b\x32\x32.tecton_proto.data.BackfillMaterializationTaskDiffR\x11\x62\x61\x63kfillTaskDiffs\x12W\n\x0f\x62\x61tch_task_diff\x18\x02 \x01(\x0b\x32/.tecton_proto.data.BatchMaterializationTaskDiffR\rbatchTaskDiff\x12Z\n\x10stream_task_diff\x18\x03 \x01(\x0b\x32\x30.tecton_proto.data.StreamMaterializationTaskDiffR\x0estreamTaskDiff\x12p\n\x1b\x62\x61\x63kfill_publish_task_diffs\x18\x04 \x03(\x0b\x32\x31.tecton_proto.data.BackfillFeaturePublishTaskDiffR\x18\x62\x61\x63kfillPublishTaskDiffs\x12m\n\x1bintegration_test_task_diffs\x18\x05 \x03(\x0b\x32..tecton_proto.data.PlanIntegrationTestTaskDiffR\x18integrationTestTaskDiffs\"\xfd\x01\n\x1e\x42\x61\x63kfillFeaturePublishTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\x12H\n\x12\x66\x65\x61ture_start_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x66\x65\x61tureStartTime\x12\x44\n\x10\x66\x65\x61ture_end_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x66\x65\x61tureEndTime\x12$\n\x0enumber_of_jobs\x18\x04 \x01(\x05R\x0cnumberOfJobs\"\xc3\x03\n\x07\x46\x63oDiff\x12\x38\n\x04type\x18\x01 \x01(\x0e\x32$.tecton_proto.data.FcoTransitionTypeR\x04type\x12\x63\n\x17transition_side_effects\x18\x0c \x01(\x0b\x32+.tecton_proto.data.FcoTransitionSideEffectsR\x15transitionSideEffects\x12?\n\rexisting_args\x18\x07 \x01(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x0c\x65xistingArgs\x12?\n\rdeclared_args\x18\x08 \x01(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x0c\x64\x65\x63laredArgs\x12\x36\n\x04\x64iff\x18\x04 \x03(\x0b\x32\".tecton_proto.data.FcoPropertyDiffR\x04\x64iff\x12Y\n\x14materialization_info\x18\x0b \x01(\x0b\x32&.tecton_proto.data.MaterializationInfoR\x13materializationInfoJ\x04\x08\t\x10\n\"C\n\x0b\x46\x63oFieldRef\x12.\n\x06\x66\x63o_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x05\x66\x63oIdJ\x04\x08\x02\x10\x03\"h\n\x11ValidationMessage\x12\x18\n\x07message\x18\x01 \x01(\tR\x07message\x12\x39\n\x08\x66\x63o_refs\x18\x02 \x03(\x0b\x32\x1e.tecton_proto.data.FcoFieldRefR\x07\x66\x63oRefs\"\x92\x01\n\x10ValidationResult\x12<\n\x06\x65rrors\x18\x01 \x03(\x0b\x32$.tecton_proto.data.ValidationMessageR\x06\x65rrors\x12@\n\x08warnings\x18\x02 \x03(\x0b\x32$.tecton_proto.data.ValidationMessageR\x08warnings\"\xa6\x02\n\x14SuccessfulPlanOutput\x12#\n\rstring_output\x18\x01 \x01(\tR\x0cstringOutput\x12\x1f\n\x0bjson_output\x18\x02 \x01(\tR\njsonOutput\x12%\n\x0e\x61pply_warnings\x18\x03 \x01(\tR\rapplyWarnings\x12(\n\x10num_fcos_changed\x18\x04 \x01(\x05R\x0enumFcosChanged\x12!\n\x0cnum_warnings\x18\x05 \x01(\x05R\x0bnumWarnings\x12T\n\x0etest_summaries\x18\x06 \x03(\x0b\x32-.tecton_proto.data.PlanIntegrationTestSummaryR\rtestSummaries\"`\n\x19IntegrationTestJobSummary\x12\x43\n\x06status\x18\x01 \x01(\x0e\x32+.tecton_proto.data.IntegrationTestJobStatusR\x06status\"\xb0\x01\n\x1aPlanIntegrationTestSummary\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12Q\n\rjob_summaries\x18\x02 \x03(\x0b\x32,.tecton_proto.data.IntegrationTestJobSummaryR\x0cjobSummaries\"\x9d\x04\n\x12StateUpdateRequest\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12\x35\n\x08\x66\x63o_args\x18\x02 \x03(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x07\x66\x63oArgs\x12R\n\x10repo_source_info\x18\x03 \x01(\x0b\x32(.tecton_proto.args.FeatureRepoSourceInfoR\x0erepoSourceInfo\x12-\n\x12suppress_recreates\x18\n \x01(\x08R\x11suppressRecreates\x12\x1f\n\x0bupgrade_all\x18\x04 \x01(\x08R\nupgradeAll\x12!\n\x0crequested_by\x18\x05 \x01(\tR\x0brequestedBy\x12R\n\x16requested_by_principal\x18\x0b \x01(\x0b\x32\x1c.tecton_proto.auth.PrincipalR\x14requestedByPrincipal\x12\x1f\n\x0bsdk_version\x18\t \x01(\tR\nsdkVersion\x12\x64\n\x15plan_integration_type\x18\x0c \x01(\x0e\x32\x30.tecton_proto.data.PlanIntegrationTestSelectTypeR\x13planIntegrationTypeJ\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\t\"\xa1\x04\n\x10StateUpdateEntry\x12\x1b\n\tcommit_id\x18\x01 \x01(\tR\x08\x63ommitId\x12\x1d\n\napplied_by\x18\x02 \x01(\tR\tappliedBy\x12S\n\x14\x61pplied_by_principal\x18\x06 \x01(\x0b\x32!.tecton_proto.auth.PrincipalBasicR\x12\x61ppliedByPrincipal\x12\x39\n\napplied_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tappliedAt\x12\x1c\n\tworkspace\x18\x04 \x01(\tR\tworkspace\x12\x1f\n\x0bsdk_version\x18\x05 \x01(\tR\nsdkVersion\x12\x39\n\ncreated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x42\n\x0bstatus_type\x18\x08 \x01(\x0e\x32!.tecton_proto.data.PlanStatusTypeR\nstatusType\x12\x16\n\x05\x65rror\x18\t \x01(\tH\x00R\x05\x65rror\x12_\n\x16successful_plan_output\x18\n \x01(\x0b\x32\'.tecton_proto.data.SuccessfulPlanOutputH\x00R\x14successfulPlanOutputB\n\n\x08response\"\xae\x03\n\x16StateUpdatePlanSummary\x12L\n\ndiff_items\x18\x01 \x03(\x0b\x32-.tecton_proto.data.StateUpdatePlanSummaryDiffR\tdiffItems\x12\x1d\n\napplied_by\x18\x02 \x01(\tR\tappliedBy\x12S\n\x14\x61pplied_by_principal\x18\x08 \x01(\x0b\x32!.tecton_proto.auth.PrincipalBasicR\x12\x61ppliedByPrincipal\x12\x1d\n\ncreated_by\x18\x03 \x01(\tR\tcreatedBy\x12\x39\n\napplied_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tappliedAt\x12\x39\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x1c\n\tworkspace\x18\x06 \x01(\tR\tworkspace\x12\x1f\n\x0bsdk_version\x18\x07 \x01(\tR\nsdkVersion\"\xa1\x03\n\x1aStateUpdatePlanSummaryDiff\x12\x19\n\x08\x66\x63o_type\x18\x01 \x01(\tR\x07\x66\x63oType\x12\x38\n\x04type\x18\x02 \x01(\x0e\x32$.tecton_proto.data.FcoTransitionTypeR\x04type\x12\x63\n\x17transition_side_effects\x18\x07 \x01(\x0b\x32+.tecton_proto.data.FcoTransitionSideEffectsR\x15transitionSideEffects\x12\x38\n\x05\x64iffs\x18\x03 \x03(\x0b\x32\".tecton_proto.data.FcoPropertyDiffR\x05\x64iffs\x12\x12\n\x04name\x18\x04 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x05 \x01(\tR\x0b\x64\x65scription\x12Y\n\x14materialization_info\x18\x06 \x01(\x0b\x32&.tecton_proto.data.MaterializationInfoR\x13materializationInfo*n\n\x11\x46\x63oTransitionType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x43REATE\x10\x01\x12\n\n\x06\x44\x45LETE\x10\x02\x12\x0b\n\x07UPGRADE\x10\x05\x12\n\n\x06UPDATE\x10\x03\x12\x0c\n\x08RECREATE\x10\x04\x12\r\n\tUNCHANGED\x10\x06*\x95\x01\n(FcoTransitionSideEffectStreamRestartType\x12\x17\n\x13RESTART_STREAM_NONE\x10\x00\x12$\n RESTART_STREAM_REUSE_CHECKPOINTS\x10\x01\x12*\n&RESTART_STREAM_CHECKPOINTS_INVALIDATED\x10\x02*\xba\x02\n\"MaterializationTaskDiffDestination\x12\x35\n1MATERIALIZATION_TASK_DIFF_DESTINATION_UNSPECIFIED\x10\x00\x12\x31\n-MATERIALIZATION_TASK_DIFF_DESTINATION_OFFLINE\x10\x01\x12\x30\n,MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE\x10\x02\x12<\n8MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE_AND_OFFLINE\x10\x03\x12:\n6MATERIALIZATION_TASK_DIFF_DESTINATION_BULK_LOAD_ONLINE\x10\x04*\xb3\x01\n\x18IntegrationTestJobStatus\x12\x1a\n\x16JOB_STATUS_UNSPECIFIED\x10\x00\x12\x1a\n\x16JOB_STATUS_NOT_STARTED\x10\x01\x12\x16\n\x12JOB_STATUS_RUNNING\x10\x02\x12\x18\n\x14JOB_STATUS_CANCELLED\x10\x03\x12\x16\n\x12JOB_STATUS_SUCCEED\x10\x04\x12\x15\n\x11JOB_STATUS_FAILED\x10\x05*D\n\x1dPlanIntegrationTestSelectType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04\x41UTO\x10\x01\x12\x08\n\x04NONE\x10\x02*\xa7\x02\n\x0ePlanStatusType\x12\x14\n\x10PLAN_UNSPECIFIED\x10\x00\x12\x10\n\x0cPLAN_CREATED\x10\x01\x12\"\n\x1ePLAN_INTEGRATION_TESTS_SKIPPED\x10\x02\x12&\n\"PLAN_INTEGRATION_TESTS_NOT_STARTED\x10\x03\x12\"\n\x1ePLAN_INTEGRATION_TESTS_RUNNING\x10\x04\x12$\n PLAN_INTEGRATION_TESTS_CANCELLED\x10\x05\x12\"\n\x1ePLAN_INTEGRATION_TESTS_SUCCEED\x10\x06\x12!\n\x1dPLAN_INTEGRATION_TESTS_FAILED\x10\x07\x12\x10\n\x0cPLAN_APPLIED\x10\x08\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.state_update__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _FCOTRANSITIONTYPE._serialized_start=5583
  _FCOTRANSITIONTYPE._serialized_end=5693
  _FCOTRANSITIONSIDEEFFECTSTREAMRESTARTTYPE._serialized_start=5696
  _FCOTRANSITIONSIDEEFFECTSTREAMRESTARTTYPE._serialized_end=5845
  _MATERIALIZATIONTASKDIFFDESTINATION._serialized_start=5848
  _MATERIALIZATIONTASKDIFFDESTINATION._serialized_end=6162
  _INTEGRATIONTESTJOBSTATUS._serialized_start=6165
  _INTEGRATIONTESTJOBSTATUS._serialized_end=6344
  _PLANINTEGRATIONTESTSELECTTYPE._serialized_start=6346
  _PLANINTEGRATIONTESTSELECTTYPE._serialized_end=6414
  _PLANSTATUSTYPE._serialized_start=6417
  _PLANSTATUSTYPE._serialized_end=6712
  _FCOPROPERTYDIFF._serialized_start=349
  _FCOPROPERTYDIFF._serialized_end=639
  _FCOTRANSITIONSIDEEFFECTS._serialized_start=642
  _FCOTRANSITIONSIDEEFFECTS._serialized_end=777
  _BATCHMATERIALIZATIONTASKDIFF._serialized_start=780
  _BATCHMATERIALIZATIONTASKDIFF._serialized_end=1010
  _STREAMMATERIALIZATIONTASKDIFF._serialized_start=1012
  _STREAMMATERIALIZATIONTASKDIFF._serialized_end=1082
  _BACKFILLMATERIALIZATIONTASKDIFF._serialized_start=1085
  _BACKFILLMATERIALIZATIONTASKDIFF._serialized_end=1428
  _PLANINTEGRATIONTESTTASKDIFF._serialized_start=1430
  _PLANINTEGRATIONTESTTASKDIFF._serialized_end=1498
  _MATERIALIZATIONINFO._serialized_start=1501
  _MATERIALIZATIONINFO._serialized_end=2028
  _BACKFILLFEATUREPUBLISHTASKDIFF._serialized_start=2031
  _BACKFILLFEATUREPUBLISHTASKDIFF._serialized_end=2284
  _FCODIFF._serialized_start=2287
  _FCODIFF._serialized_end=2738
  _FCOFIELDREF._serialized_start=2740
  _FCOFIELDREF._serialized_end=2807
  _VALIDATIONMESSAGE._serialized_start=2809
  _VALIDATIONMESSAGE._serialized_end=2913
  _VALIDATIONRESULT._serialized_start=2916
  _VALIDATIONRESULT._serialized_end=3062
  _SUCCESSFULPLANOUTPUT._serialized_start=3065
  _SUCCESSFULPLANOUTPUT._serialized_end=3359
  _INTEGRATIONTESTJOBSUMMARY._serialized_start=3361
  _INTEGRATIONTESTJOBSUMMARY._serialized_end=3457
  _PLANINTEGRATIONTESTSUMMARY._serialized_start=3460
  _PLANINTEGRATIONTESTSUMMARY._serialized_end=3636
  _STATEUPDATEREQUEST._serialized_start=3639
  _STATEUPDATEREQUEST._serialized_end=4180
  _STATEUPDATEENTRY._serialized_start=4183
  _STATEUPDATEENTRY._serialized_end=4728
  _STATEUPDATEPLANSUMMARY._serialized_start=4731
  _STATEUPDATEPLANSUMMARY._serialized_end=5161
  _STATEUPDATEPLANSUMMARYDIFF._serialized_start=5164
  _STATEUPDATEPLANSUMMARYDIFF._serialized_end=5581
# @@protoc_insertion_point(module_scope)



import logging
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Tuple

import click
import yaml
from colorama import Fore

from tecton import version
from tecton._internals.tecton_pydantic import pydantic_v1
from tecton.cli import printer
from tecton.cli.command import TectonGroup
from tecton.framework import repo_config as repo_config_module
from tecton_core import repo_file_handler


logger = logging.getLogger(__name__)

DEFAULT_REPO_CONFIG_NAME = "repo.yaml"

# Templated starter repo config string. Simpler than bundling as a data asset.
_STARTER_REPO_CONFIG = """# This is the Tecton repo config. It's used to configure how Tecton builds and applies your
# feature definitions during `tecton plan/apply/test`.
#
# By default, the Tecton CLI will use the Repo Config specified at <TECTON_REPO_ROOT>/repo.yaml, but you
# can specify another file by using `tecton plan --config my_config.yaml`.

# The `defaults` keyword specifies default parameter values for Tecton objects defined in your Feature Repository.
# For example, you can set a default `tecton_materialization_runtime` for all Batch Feature Views.
# Defaults can be overridden on a per-object basis in your Python feature definitions.
# See Tecton's documentation for details on which Tecton objects are currently supported by the `defaults` keyword.

defaults:
  batch_feature_view:
    tecton_materialization_runtime: {current_version}
    environment: {environment}  # For Rift-based Batch Feature Views
  stream_feature_view:
    tecton_materialization_runtime: {current_version}
    environment: {environment}  # For Rift-based Stream Feature Views
  feature_table:
    tecton_materialization_runtime: {current_version}

# Below is an example of other defaults that can be set using the `defaults` keyword.
# defaults:
#   batch_feature_view:
#     tecton_materialization_runtime: {current_version}
#     online_store:
#       kind: RedisConfig
#     offline_store:
#       kind: OfflineStoreConfig
#       staging_table_format:
#         kind: ParquetConfig
#     batch_compute:
#       kind: DatabricksClusterConfig
#       instance_type: m5.xlarge
#       number_of_workers: 2
#       extra_pip_dependencies:
#         - haversine==2.8.0
#   stream_feature_view:
#     tecton_materialization_runtime: {current_version}
#     stream_compute:
#       kind: DatabricksClusterConfig
#       instance_availability: on_demand
#       instance_type: m5.2xlarge
#       number_of_workers: 4
#     offline_store:
#       kind: OfflineStoreConfig
#       staging_table_format:
#         kind: ParquetConfig
#   feature_table:
#     tecton_materialization_runtime: {current_version}
#     batch_compute:
#       kind: DatabricksClusterConfig
#       instance_type: m5.xlarge
#       number_of_workers: 2
#     online_store:
#       kind: RedisConfig
#   feature_service:
#     on_demand_environment: tecton-python-extended:0.4
"""


@click.command("repo-config", cls=TectonGroup)
def repo_config_group():
    """Create, inspect, or debug the repo configuration."""


@repo_config_group.command("show")
@click.argument(
    "config", required=False, default=None, type=click.Path(exists=True, dir_okay=False, path_type=Path, readable=True)
)
def show(config: Optional[Path]):
    """Print out the parsed repo config at CONFIG path. Defaults to the repo.yaml at the repo root."""
    if config is None:
        config = get_default_repo_config_path()

    printer.safe_print(f"Loading and printing repo config at path: {config}")
    load_repo_config(config)
    loaded_config = repo_config_module.get_repo_config()

    # TODO(jake): This prints out formatted JSON, which is decent. Using a library like "rich" or a custom function
    # to print out the model would be better.
    printer.safe_print(loaded_config.json(exclude_unset=True, indent=4, by_alias=True))


@repo_config_group.command("init")
@click.argument("config", required=False, default=None, type=click.Path(exists=False, path_type=Path, readable=True))
def init(config: Optional[Path]):
    """Write out a starter repo config to the provided CONFIG path. Default path is REPO_ROOT/repo.yaml."""
    if config is None:
        config = get_default_repo_config_path()

    if config.exists():
        printer.safe_print(Fore.RED + f"A file already exists at {config}. Aborting." + Fore.RESET)
        sys.exit(1)

    create_starter_repo_config(config_path=config)
    printer.safe_print(
        f"Starter repo config written to {config}.\n\n💡 We recommend tracking this file in git.", file=sys.stderr
    )


def create_starter_repo_config(config_path: Path):
    """Create a starter repo config to config_path."""
    sdk_version = version.get_version()

    # TODO (TEC-19058): hard-coded pending a process that publishes the environments at the same time as the SDK
    environment = "tecton-rift-core-0.9.9" if sdk_version != "99.99.99" else ""
    formatted_start_repo_config = _STARTER_REPO_CONFIG.format(current_version=sdk_version, environment=environment)

    with open(config_path, "w") as file:
        file.write(formatted_start_repo_config)


def get_default_repo_config_path() -> Path:
    repo_file_handler.ensure_prepare_repo()
    repo_root = repo_file_handler.repo_root()
    return Path(repo_root) / DEFAULT_REPO_CONFIG_NAME


def load_repo_config(repo_config_path: Path) -> None:
    """Load the repo config from the yaml file at repo_config_path."""
    if not repo_config_path.exists():
        printer.safe_print(
            Fore.RED
            + f"A repo config is required for this command and not found at {repo_config_path}. Run `tecton repo-config init` to create a starter config."
            + Fore.RESET
        )
        sys.exit(1)

    # Existence of the repo config path should already be validated by this point.
    with open(repo_config_path, "r") as file:
        try:
            repo_config_dict = yaml.safe_load(file) or {}
        except yaml.YAMLError as e:
            printer.safe_print(
                Fore.RED
                + f"Failed to parse the Tecton repo config at {repo_config_path}. Likely invalid YAML."
                + f"\nError details:\n{e}"
                + Fore.RESET
            )
            sys.exit(1)

    try:
        config = repo_config_module.RepoConfig(**repo_config_dict)
    except pydantic_v1.ValidationError as e:
        printer.safe_print(
            Fore.RED + f"Invalid Tecton repo config at {repo_config_path}.\n\n" + "Error details:" + Fore.RESET
        )

        for error_details in e.errors():
            pretty_details = _PrettyPydanticErrorDetails.from_pydantic_details(error_details)
            # Include the problem value at the error path. Note that in Pydantic V2, the error includes the input value,
            # so we can remove this logic when moving to Pydantic V2.
            input_value = _get_nested_value(repo_config_dict, pretty_details.raw_path)
            printer.safe_print(
                Fore.RED
                + f"  {pretty_details.message}\n"
                + f"    Path: '{pretty_details.path}'\n"
                + f"    Value: {input_value} ({type(input_value)})"
                + Fore.RESET
            )

        logger.debug(f"Raw repo config error: {e}.")
        sys.exit(1)

    repo_config_module.set_repo_config(config)
    logger.debug(f"Successfully loaded repo config: {config}")


@dataclass
class _PrettyPydanticErrorDetails:
    """Used to improve Pydantic repo config error rendering."""

    raw_path: Tuple[str, ...]
    raw_message: str
    value: Any
    error_type: str

    @classmethod
    def from_pydantic_details(cls, details: Mapping) -> "_PrettyPydanticErrorDetails":
        return cls(
            raw_path=details.get("loc", ()),
            raw_message=details.get("msg", "Unknown error."),
            value=details.get("input"),
            error_type=details.get("type", "unknown_error"),
        )

    @property
    def path(self) -> Optional[str]:
        return ".".join(self.raw_path)

    @property
    def field_name(self) -> str:
        if len(self.raw_path) == 0:
            msg = "Invalid path provided for exception."
            raise ValueError(msg)
        return self.raw_path[-1]

    @property
    def message(self) -> str:
        pretty_error_msgs = {
            "value_error.extra": f"Unexpected field name '{self.field_name}' found at path",
            "value_error.discriminated_union.missing_discriminator": f"Missing 'kind' field at '{self.path}'. Kind is needed to discrimate Union types, e.g. 'kind: ParquetConfig'",
        }
        return pretty_error_msgs.get(self.error_type, self.raw_message)


def _get_nested_value(dct: dict, raw_path: Tuple[str, ...]) -> Any:
    """Safely traverse a nested dictionary using a tuple of keys. Return the value at path."""
    current = dct
    for key in raw_path:
        if not isinstance(current, dict):
            # Not at a leaf in the path, but the current value is not a dictionary. This is unexpected, but return
            # None to be safe.
            logger.warning(f"Expected value at nested path {raw_path} but did not find one.")
            return None

        is_discriminator_key = "kind" in current and current["kind"] == key
        if is_discriminator_key:
            # For some reason Pydantic V1 includes Union discriminators in the path, e.g. DatabricksClusterConfig in
            # 'defaults.batch_feature_view.batch_compute.DatabricksClusterConfig.number_of_workers'. Skip these path
            # entries.
            continue

        if key not in current:
            return None

        current = current[key]

    return current
